from typing import List, Literal, Tuple
import numpy as np
from optbinning import MulticlassOptimalBinning
import pandas as pd
from scipy.stats import variation, kstest
from luntaiDs.ModelingTools.Explore.engines.base import _BaseEDAEngine, _BaseNumericHelper, serialize
from luntaiDs.ModelingTools.Explore.summary import BinaryStatAttr, BinaryStatSummary, CategStatAttr, \
    CategStatSummary, CategUniVarClfTargetCorr, DescStat, NumericUniVarClfTargetCorr, QuantileStat, StatVar, XtremeStat, NumericStatAttr, \
        NumericStatSummary, exp10pc, log10pc


class NumericHelperPd(_BaseNumericHelper):
    def __init__(self, vector: pd.Series):
        self._vector = vector.astype("float")
    
    def get_descriptive_stat(self) -> DescStat:
        stat_descriptive = DescStat(
            mean=serialize(self._vector.mean()),
            var=serialize(self._vector.var()),
            std=serialize(self._vector.std()),
            skew=serialize(self._vector.skew()),
            kurt=serialize(self._vector.kurtosis()),
            mad=serialize((self._vector - self._vector.mean()).abs().mean()), # mean absolute deviation
            cv=serialize(variation(self._vector, nan_policy="omit")),
            normality_p=serialize(kstest(self._vector.dropna(), cdf="norm")[1]),
        )
        return stat_descriptive
    
    def get_quantile_stat(self) -> QuantileStat:
        stat_quantile = QuantileStat(
            minimum=serialize(self._vector.min()),
            perc_1th=serialize(self._vector.quantile(0.01)),
            perc_5th=serialize(self._vector.quantile(0.05)),
            q1=serialize(self._vector.quantile(0.25)),
            median=serialize(self._vector.quantile(0.5)),
            q3=serialize(self._vector.quantile(0.75)),
            perc_95th=serialize(self._vector.quantile(0.95)),
            perc_99th=serialize(self._vector.quantile(0.99)),
            maximum=serialize(self._vector.max()),
            iqr=serialize(self._vector.quantile(0.75) - self._vector.quantile(0.25)),
            range=serialize(self._vector.max() - self._vector.min()),
        )
        return stat_quantile        
        
    
    def get_xtreme_stat(self, xtreme_method: Literal["iqr", "quantile"] = "iqr") -> XtremeStat:
        if xtreme_method == "iqr":
            q1 = self._vector.quantile(0.25)
            q3 = self._vector.quantile(0.75)
            iqr = q3 - q1
            lbound = q1 - 1.5 * iqr
            rbound = q3 + 1.5 * iqr

        elif xtreme_method == "quantile":
            lbound = self._vector.quantile(0.01)
            rbound = self._vector.quantile(0.99)
        else:
            raise ValueError("xtreme_method can only be iqr or quantile")

        lvector = self._vector[self._vector < lbound]
        rvector = self._vector[self._vector > rbound]

        return XtremeStat(
            lbound=serialize(lbound),
            rbound=serialize(rbound),
            lxtreme_mean=serialize(lvector.mean()),
            lxtreme_median=serialize(lvector.median()),
            rxtreme_mean=serialize(rvector.mean()),
            rxtreme_median=serialize(rvector.median()),
        )
        
    def get_histogram(self, n_bins:int) -> Tuple[np.ndarray, np.ndarray]:
        hist_, bin_edges_ = np.histogram(self._vector, bins=n_bins)
        return hist_, bin_edges_

def _combine_x_y(x: pd.Series, y: pd.Series, 
        combine_x_categ:bool = False) -> pd.DataFrame:
    df = pd.DataFrame({'x' : x.values, 'y' : y.values})
    df = df.dropna(
        subset = ['y'], 
        axis = 0, 
        how = 'any'
    )
    # categorical variable for x only:
    if combine_x_categ:
        vc = df['x'].value_counts(normalize = True)
        others = vc.iloc[20:].index  # only keep top 20 categories
        df.loc[df['x'].isin(others), 'x'] = 'Others'
    
    return df


class EDAEnginePandas(_BaseEDAEngine):
    def __init__(self, df: pd.DataFrame):
        self._df = df
        
    def get_columns(self) -> List[str]:
        """get all column list from given dataset

        :return List[str]: list of columns in the dataset
        """
        return self._df.columns.tolist()
        
    def _fit_common_categ(self, colname: str, attr: CategStatAttr) -> CategStatSummary:
        """common categorical variable fitting can be reused by subclass categorical fitting

        :param str colname: column name
        :param CategStatAttr attr: categorical variable attribute object
        :return CategStatSummary: categorical variable summary object
        """
        vector = self._df[colname]
        
        # total number of records
        total_ = len(vector)
        # missing rate
        missing = vector.isna()
        missing_ = StatVar(
            value=serialize(missing.sum()), 
            perc=serialize(missing.mean())
        )
        # num of unique value (dropped NA value)
        unique = len(vector.dropna().unique())
        unique_ = StatVar(
            value=serialize(unique), 
            perc=serialize(unique / total_)
        )
        # category counts without null
        vcounts_ = (
            vector
            .value_counts(dropna=True, normalize=False)
            .fillna(0)
        )
        vpercs_ = (
            vcounts_ / vcounts_.sum()
        ).fillna(0)
        
        return CategStatSummary(
            colname_ = colname,
            total_ = total_,
            missing_ = missing_,
            unique_ = unique_,
            vcounts_ = vcounts_,
            vpercs_ = vpercs_
        )
        
    def fit_binary(self, colname: str, attr: BinaryStatAttr) -> BinaryStatSummary:
        """binary categorical variable fitting

        :param str colname: column name
        :param BinaryStatAttr attr: binary variable attribute object
        :return BinaryStatSummary: binary variable summary object
        """
        vector = self._df[colname]
        
        # reuse common stat
        common_stat_summary = self._fit_common_categ(colname, attr)
        #num_unique = common_stat_summary.unique_.value
        #assert num_unique <= 2, f"Labeled as binary value, but {num_unique} categories are found"
        
        # binary stat
        vector_binary = vector.copy()
        vector_binary[~vector_binary.isna()] = vector_binary[~vector_binary.isna()].apply(
            lambda v: 1 if v in attr.pos_values_ else 0
        )
        vector_binary = vector_binary.fillna(int(attr.na_to_pos_))
        binary_vcounts_ = (
            vector_binary
            .value_counts(dropna=True, normalize=False)
            .fillna(0)
        )
        binary_vpercs_ = (
            binary_vcounts_ / binary_vcounts_.sum()
        ).fillna(0)
        
        return BinaryStatSummary(
            colname_ = common_stat_summary.colname_,
            total_ = common_stat_summary.total_,
            missing_ = common_stat_summary.missing_,
            unique_ = common_stat_summary.unique_,
            vcounts_ = common_stat_summary.vcounts_,
            vpercs_ = common_stat_summary.vpercs_,
            binary_vcounts_ = binary_vcounts_,
            binary_vpercs_ = binary_vpercs_
        )
        
    def fit_numeric(self, colname: str, attr: NumericStatAttr) -> NumericStatSummary:
        """numeric variable fitting

        :param str colname: column name
        :param NumericStatAttr attr: numeric variable attribute object
        :return NumericStatSummary: numeric variable summary object
        """
        vector = self._df[colname].astype('float')
        
        # total number of records
        total_ = len(vector)
        # missing rate
        missing = vector.isna()
        missing_ = StatVar(
            value=serialize(missing.sum()), 
            perc=serialize(missing.mean())
        )
        # zero rate
        zeros = vector[vector == 0].count()
        zeros_ = StatVar(
            value=serialize(zeros), 
            perc=serialize(zeros / total_)
        )
        
        # infinite value
        infs_pos = vector[vector == np.inf].count()
        infs_pos_ = StatVar(
            value=serialize(infs_pos), 
            perc=serialize(infs_pos / total_)
        )
        infs_neg = vector[vector == -np.inf].count()
        infs_neg_ = StatVar(
            value=serialize(infs_neg), 
            perc=serialize(infs_neg / total_)
        )
        
        vector = vector.dropna()
        vector = vector[(vector < np.inf) & (vector > -np.inf)]
        if attr.setaside_zero_:
            vector = vector[vector != 0]
            
        # xtreme value
        if attr.xtreme_method_:
            xst = NumericHelperPd(vector = vector)
            xtreme_stat_ = xst.get_xtreme_stat(
                xtreme_method=attr.xtreme_method_
            )
            vector_clean = vector[
                (vector >= xtreme_stat_.lbound) 
                & (vector <= xtreme_stat_.rbound)
            ]

        else:
            vector_clean = vector.copy()
            xtreme_stat_ = None
            
        num_xtreme = len(vector) - len(vector_clean)
        xtreme_ = StatVar(
            value=serialize(num_xtreme), 
            perc=serialize(num_xtreme / total_)
        )
        
        # percentile & descriptive statistics
        xst_clean = NumericHelperPd(vector = vector_clean)
        stat_quantile_ = xst_clean.get_quantile_stat()
        stat_descriptive_ = xst_clean.get_descriptive_stat()
        
        # histogram
        hist_, bin_edges_ = xst_clean.get_histogram(n_bins=attr.bins_)
        
        num_stat = NumericStatSummary(
            colname_ = colname,
            total_ = total_,
            missing_ = missing_,
            zeros_ = zeros_,
            infs_pos_ = infs_pos_,
            infs_neg_ = infs_neg_,
            stat_quantile_ = stat_quantile_,
            stat_descriptive_ = stat_descriptive_,
            hist_ = hist_,
            bin_edges_ = bin_edges_,
            xtreme_ = xtreme_,
            xtreme_stat_ = xtreme_stat_
        )
        
        if attr.log_scale_:
            vector_log = pd.Series(
                log10pc(vector), 
                name=colname, 
                dtype="float"
            )

            # xtreme value
            if attr.xtreme_method_:
                xst_log = NumericHelperPd(vector = vector_log)
                num_stat.xtreme_stat_log_ = xst_log.get_xtreme_stat(xtreme_method=attr.xtreme_method_)
                vector_clean_log = vector_log[
                    (vector_log >= num_stat.xtreme_stat_log_.lbound)
                    & (vector_log <= num_stat.xtreme_stat_log_.rbound)
                ]
            else:
                vector_clean_log = vector_log.copy()

            num_xtreme_log = len(vector_log) - len(vector_clean_log)
            num_stat.xtreme_log_ = StatVar(
                value=serialize(num_xtreme_log), 
                perc=serialize(num_xtreme_log / total_)
            )
            # percentile & descriptive statistics
            xstat_clean_log = NumericHelperPd(vector = vector_clean_log)
            
            num_stat.stat_quantile_log_ = xstat_clean_log.get_quantile_stat()
            num_stat.stat_descriptive_log_ = xstat_clean_log.get_descriptive_stat()
            # histogram
            num_stat.hist_log_, num_stat.bin_edges_log_ = xstat_clean_log.get_histogram(n_bins=attr.bins_)
            
        return num_stat
    
    def fit_univarclf_categ(self, x_col: str, y_col: str) -> CategUniVarClfTargetCorr:
        """fit univariate correlation for classififer target and categorical variable

        :param str x_col: column of a categorical variable
        :param str y_col: column of the classifier target varaible
        :return CategUniVarClfTargetCorr: result object
        """
        
        df = _combine_x_y(
            self._df[x_col].astype('str').fillna('Missing_'), # need to fill as separate categ
            self._df[y_col].astype('str'), 
            combine_x_categ = True # combine x's categories to be less than 20
        )
        
        # categorical:  p(x | y)
        p_x_y_ = {}
        for y in df['y'].unique():
            cs = (
                df
                .loc[df['y'] == y, ['x']]
                .groupby('x')
                .size()
            )
            d = pd.DataFrame({
                    'count' : cs.values, 
                    'perc' : cs  / cs.sum()
                }, 
                index = cs.index
            )
            p_x_y_[y] = d.fillna(0) #.to_dict(orient = 'index')
            
        # categorical:  p(y | x)  prob (event rate when binary clf) by category
        p_y_x_ = {}
        for x in df['x'].unique():
            cs = (
                df
                .loc[df['x'] == x, ['y']]
                .groupby('y')
                .size()
            )
            d = pd.DataFrame({
                    'count' : cs.values, 
                    'perc' : cs  / cs.sum()
                }, 
                index = cs.index
            )
            p_y_x_[x] = d.fillna(0) #.to_dict(orient = 'index')
            
        return CategUniVarClfTargetCorr(
            colname_ = x_col,
            yname_ = y_col,
            ylabels_ = (
                self._df[y_col]
                .dropna()
                .unique()
                .astype('str')
                .tolist()
            ),
            p_x_y_ = p_x_y_,
            p_y_x_ = p_y_x_
        )
    
    def fit_univarclf_numeric(self, x_col: str, y_col: str) -> NumericUniVarClfTargetCorr:
        """fit univariate correlation for classififer target and numeric variable

        :param str x_col: column of a numeric variable
        :param str y_col: column of the classifier target varaible
        :return NumericUniVarClfTargetCorr: result object
        """
        df = _combine_x_y(
            self._df[x_col].astype('float'), 
            self._df[y_col].astype('str'), 
        )
        
        # numerical:  p(x | y)  distribution by target (boxplot)
        t = {'origin' : {}, 'log' : {}}
        for y in df['y'].unique():
            # EDA on given subset
            eda = EDAEnginePandas(df[df['y'] == y])
            summary = eda.fit_numeric(
                colname = 'x', # already renamed
                attr = NumericStatAttr(
                    setaside_zero_ = False, 
                    log_scale_ = True, 
                    xtreme_method_ = 'iqr', 
                    bins_ = 100
                )
            )
            
            t['origin'][y] = {
                'lbound': summary.xtreme_stat_.lbound,
                'q1' : summary.stat_quantile_.q1,
                'mean' : summary.stat_descriptive_.mean,
                'median' : summary.stat_quantile_.median,
                'q3' : summary.stat_quantile_.q3,
                'rbound': summary.xtreme_stat_.rbound,
            }
            t['log'][y] = {
                'lbound': summary.xtreme_stat_log_.lbound,
                'q1' : summary.stat_quantile_log_.q1,
                'mean' : summary.stat_descriptive_log_.mean,
                'median' : summary.stat_quantile_log_.median,
                'q3' : summary.stat_quantile_log_.q3,
                'rbound': summary.xtreme_stat_log_.rbound,
            }
            
        p_x_y_ = {
            'origin' : pd.DataFrame.from_dict(t['origin'], orient='index'),
            'log' : pd.DataFrame.from_dict(t['log'], orient='index')
        }
            
            
        # numerical:  p(y | x)  prob (event rate if binary clf) by bucketized x
        mob = MulticlassOptimalBinning(
            prebinning_method = 'quantile',
            monotonic_trend = 'auto',
            max_n_prebins = 20,
            min_prebin_size = 0.05,
        )
        mob.fit(x = df['x'], y = df['y'])
        p_y_x_ = mob.binning_table.build(add_totals=False)
        
        return NumericUniVarClfTargetCorr(
            colname_ = x_col,
            yname_ = y_col,
            ylabels_ = (
                self._df[y_col]
                .dropna()
                .unique()
                .astype('str')
                .tolist()
            ),
            p_x_y_ = p_x_y_,
            p_y_x_ = p_y_x_
        )

    



import json
from pathlib import Path
from typing import List
from fsspec import AbstractFileSystem

class _BaseModelRegistry:
    """Base class for Simple Model Registry
    model versioning is achieved by tracking model_id and its configuration

    configuration format:
    {
        "prod" : "MODEL_ID",
        "archive" : {
            "MODEL_ID_A" : {CONFIG_A},
            "MODEL_ID_B" : {CONFIG_B},
        }
    }
    """

    def load_config(self) -> dict:
        """load configuration
        
        :return dict: configuration in dictionary format
        """
        raise NotImplementedError("")

    def save_config(self, config:dict):
        """save configuration dictionary

        :param dict config: configuration in dictionary format
        """
        raise NotImplementedError("")

    def get_prod_model_id(self) -> str:
        """get deployed (prod) model id

        :return str: model id that is deployed
        """
        return self.load_config().get("prod")

    def get_model_list(self) -> List[str]:
        """get a list of model ids

        :return List[str]: list of model ids registered
        """
        return list(self.load_config().get("archive").keys())

    def get_model_config(self, model_id: str) -> dict:
        """get specific configuration part for a given model id

        :param str model_id: model id of the model
        :return dict: the specific configuration part
        """
        assert model_id in self.get_model_list(), "Model Id not found in registry, please register one first"
        return self.load_config().get("archive").get(model_id)
    
    def upsert_model_config(self, model_id: str, model_config: dict):
        """update or insert (if not exist) model configuration

        :param str model_id: model id of the model
        :param dict model_config: the specific configuration part
        """
        config = self.load_config()
        config['archive'][model_id] = model_config
        self.save_config(config)
        
    def delete_model_config(self, model_id: str):
        """delete model configuration

        :param str model_id: model id of the model
        """
        config = self.load_config()
        config['archive'].pop(model_id)
        if config['prod'] == model_id:
            config['prod'] = None
        self.save_config(config)

    def load_model(self, model_id: str):
        """loading the model into memory

        :param str model_id: the model id for the model
        :return _type_: model object into memory
        """
        config = self.get_model_config(model_id)
        return self.load_model_by_config(config)

    def load_prod(self):
        """load deployed model into memory

        :return _type_: deployed model object into memory
        """
        return self.load_model(self.get_prod_model_id())

    def register(self, model_id: str, *args, **kws):
        """register your model into the system

        :param str model_id: your model id
        """
        assert model_id not in self.get_model_list(), "Model Id already registered, please try another one"
        model_config = self.save_model_and_generate_config(model_id, *args, **kws)
        self.upsert_model_config(model_id, model_config)

    def remove(self, model_id: str):
        """delete model from registry

        :param str model_id: model id to be removed
        """
        assert model_id in self.get_model_list(), "Model Id not found"

        self.delete_model_files(model_id)
        self.delete_model_config(model_id)

    def deploy(self, model_id: str):
        """deploy selected model to prod, simply achieved by setting it to prod field

        :param str model_id: model id to be deployed
        """
        # check if model id exists
        assert model_id in self.get_model_list(), "Model Id not found in registry, please register one first"
        config = self.load_config()
        config["prod"] = model_id
        self.save_config(config)

    def delete_model_files(self, model_id: str):
        """delete model related files/data

        :param str model_id: model id to be deleted
        """
        raise NotImplementedError("")

    def load_model_by_config(self, config: dict):
        """load the model using configuration file

        :param dict config: configuration for the model
        """
        raise NotImplementedError("")

    def save_model_and_generate_config(self, model_id:str, *args, **kws) -> dict:
        """save model and generate configuration for this model

        :param str model_id: model id to be generated
        :return dict: the configuration generated by creating model into the system
        """
        raise NotImplementedError("")
    
    
class _BaseModelRegistryFileSystem(_BaseModelRegistry):
    """Base class for Simple Model Registry on fsspec compatible Filesystem
    Note only the configuration will be saved on the given file system, not model data
    model versioning is achieved by tracking model_id and its configuration

    configuration format:
    {
        "prod" : "MODEL_ID",
        "archive" : {
            "MODEL_ID_A" : {CONFIG_A},
            "MODEL_ID_B" : {CONFIG_B},
        }
    }
    """
    def __init__(self, fs: AbstractFileSystem, config_js_path: str):
        """model registry where config is saved to given filesystem

        :param AbstractFileSystem fs: the fsspec compatible filesystem
        :param str config_js_path: path of the js config file, if on object storage, 
            the full path including buckets
        """
        self._fs = fs
        self._config_js_path = config_js_path
        self._fs.makedirs(
            path = Path(config_js_path).parent.as_posix(),
            exist_ok = True
        )

    def load_config(self) -> dict:
        """load configuration
        
        :return dict: configuration in dictionary format
        """
        if self._fs.exists(self._config_js_path):
            with self._fs.open(self._config_js_path, 'r') as obj:
                config = json.loads(obj.read())
        else:
            config = {
                "prod" : None,
                "archive": {}
            }
        return config

    def save_config(self, config:dict):
        """save configuration dictionary

        :param dict config: configuration in dictionary format
        """
        with self._fs.open(self._config_js_path, 'w') as obj:
            json.dump(config, obj, indent = 4)


import ibis
import pandas as pd
from luntaiDs.CommonTools.dbapi import baseDbInf
from luntaiDs.CommonTools.dtyper import DSchema

class BaseWarehouseHandler:
    @classmethod
    def connect(cls, db_conf: baseDbInf, **settings):
        cls._db_conf = db_conf
        
    def is_exist(self, schema: str, table: str) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created
        
        :param str schema: schema/database
        :param str table: table name
        :return bool: whether the table and schema exists and ready
        """
        raise NotImplementedError("")
    
    def get_table(self, schema: str, table: str) -> ibis.expr.types.Table:
        """get the ibis table
        
        :param str schema: schema/database
        :param str table: table name
        """
        raise NotImplementedError("")
    
    def create_schema(self, schema: str):
        """create database/schema
        
        :param str schema: schema/database
        """
        raise NotImplementedError("")
        
    def create_table(self, schema: str, table: str, col_schemas: DSchema, **settings):
        """initialize/create table in the underlying data warehouse system

        :param str schema: schema/database
        :param str table: table name
        :param DSchema col_schemas: data column schema
        :param List[str] primary_keys: primary keys, defaults to None
        """
        raise NotImplementedError("")
            
    def get_dtypes(self, schema: str, table: str) -> ibis.Schema:
        """Check dtypes of the given schema/dataset

        :return:
        """
        raise NotImplementedError("")
        
    def query(self, sql: str, schema: str = None) -> ibis.expr.types.Table:
        """query using SQL

        :return: ibis dataframe
        """
        raise NotImplementedError("")
    
    def execute(self, sql: str, **kws):
        """execute SQL for operations like insert/update/delete

        :return: ibis dataframe
        """
        raise NotImplementedError("")
        
    def save_pandas(self, df: pd.DataFrame, schema: str, table: str, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: pandas table
        :param str schema: schema/database
        :param str table: table name
        """
        raise NotImplementedError("")
    
    def save_ibis(self, df: ibis.expr.types.Table, schema: str, table: str, **kws):
        """The pure logic to save ibis dataframe to the system, without handling existing record problem

        :param ibis.expr.types.Table df: ibis table
        :param str schema: schema/database
        :param str table: table name
        """
        raise NotImplementedError("")
        
    def delete_table(self, schema: str, table: str):
        """drop whole table
        
        :param str schema: schema/database
        :param str table: table name
        """
        raise NotImplementedError("")
    
    def truncate_table(self, schema: str, table: str):
        """Truncate (remove) data from table
        
        :param str schema: schema/database
        :param str table: table name
        """
        raise NotImplementedError("")


from typing import Tuple

import pandas as pd
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from itertools import cycle
from scipy import stats
from sklearn.metrics import confusion_matrix, precision_recall_curve, roc_curve, classification_report, auc, \
    roc_auc_score, average_precision_score
from sklearn.model_selection import cross_val_score, cross_val_predict, train_test_split
from sklearn.preprocessing import label_binarize


class Validator:
    pass

class ClfValidator(Validator):
    def __init__(self, clf, Xs, y_actual, y_labels=None, binary:bool = True):
        self.binary = binary
        self.clf = clf
        self.Xs = Xs
        self.y_actual = y_actual
        self.y_predict_cv = clf.predict(Xs)
        self.y_scores = clf.predict_proba(Xs)

        # map y labels
        if y_labels is None:
            self.y_labels = clf.classes_
        else:
            self.y_labels = y_labels

        self.y_actual_binary = label_binarize(self.y_actual, classes=self.y_labels)
        if len(self.y_labels) == 2 or binary:  # adjust for binary classifier -- increase to 2D dimension
            self.y_actual_binary = np.c_[1 - self.y_actual_binary, self.y_actual_binary]

    def confusion_matrix(self, plot: bool = True, figsize=(12, 5)):

        cm = confusion_matrix(self.y_actual, self.y_predict_cv)
        cm_sum = np.sum(cm, axis=1, keepdims=True)
        cm_perc = cm / cm_sum.astype(float)
        np.fill_diagonal(cm_perc, 0)  # fill diagonal value for error chart to zero

        cm = pd.DataFrame(cm, index=self.y_labels, columns=self.y_labels)
        cm.index.name = 'Actual'
        cm.columns.name = 'Predicted'
        cm_perc = pd.DataFrame(cm_perc, index=self.y_labels, columns=self.y_labels)
        cm_perc.index.name = 'Actual'
        cm_perc.columns.name = 'Predicted'

        if plot:
            annot1 = np.empty_like(cm).astype(str)  # absolute value chart (colorful heatmap)
            annot2 = np.empty_like(cm).astype(str)  # error chart (gray scale)
            nrows, ncols = cm.shape
            for i in range(nrows):
                for j in range(ncols):
                    c = cm.iloc[i, j]
                    p = cm_perc.iloc[i, j]
                    if c == 0:
                        annot1[i, j] = ''
                        annot2[i, j] = ''
                    else:
                        annot1[i, j] = f'{c : .0f}'

                        if i != j:
                            annot2[i, j] = f'{p :.0%}'
                        else:
                            annot2[i, j] = ''

            fig, axes = plt.subplots(1, 2, sharex=True, figsize=figsize)
            sns.heatmap(ax=axes[0], data=cm, annot=annot1, fmt='')
            sns.heatmap(ax=axes[1], data=cm_perc, annot=annot2, fmt='', cmap=plt.cm.gray, vmin=0, vmax=1)

            axes[0].set_title('Confusion Matrix')
            axes[1].set_title('Error Heatmap - p(predict | actual)')
        return cm, cm_perc

    def average_precision_score(self, average: str = "macro") -> float:
        """

        :param average: {‘micro’, ‘macro’}, default='macro'
        """
        return average_precision_score(self.y_actual_binary, self.y_scores, average = average)

    def precision_recall(self, average: str = "macro", plot: bool = True, figsize = (5, 5)) -> dict:
        """

        :param average: {‘micro’, ‘macro’}, default='macro'
        :param plot: whether to plot or not
        :return: a dictionary {label : df[threshold, precision, recall]}
        """
        results = dict()
        aps = dict()

        for i, label in enumerate(self.y_labels):
            precision, recall, threshold = precision_recall_curve(self.y_actual_binary[:, i], self.y_scores[:, i])
            results[label] = pd.DataFrame({
                'threshold': threshold,
                'precision': precision[:-1],
                'recall': recall[:-1]
            })
            aps[label] = auc(recall, precision)

        # average Precision, Recall (micro/macro)
        if average == 'micro':
            ## micro ROC
            pre_avg, rec_avg, threshold_avg = precision_recall_curve(self.y_actual_binary.ravel(), self.y_scores.ravel())
        elif average == 'macro':
            ## macro ROC
            all_prec = np.unique(np.concatenate([results[label]['precision'] for label in self.y_labels]))

            #### Then interpolate all ROC curves at this points
            mean_rec = np.zeros_like(all_prec)
            for label in self.y_labels:
                mean_rec += np.interp(all_prec, results[label]['precision'], results[label]['recall'])

            mean_rec /= len(self.y_labels)

            pre_avg = all_prec
            rec_avg = mean_rec
            threshold_avg = np.empty(pre_avg.shape[0] - 1)
            threshold_avg[:] = np.nan
        else:
            raise ValueError("parameter 'average' can only be micro or macro")

        results['AVG'] = pd.DataFrame({
            'threshold': threshold_avg,
            'precision': pre_avg[:-1],
            'recall': rec_avg[:-1]
        })

        precision_avg = self.average_precision_score(average=average)
        precision_overall = self.average_precision_score(average='weighted')

        if plot:
            sns.set_style("ticks", {'axes.grid': True})
            fig, axes = plt.subplots(1, 1, figsize=figsize)
            for i, label in enumerate(self.y_labels):
                legend = f"{label} (AP = {aps[label]:.3f})"
                sns.lineplot(x = results[label]['recall'], y = results[label]['precision'], label = legend, linewidth=2)

            # average Precision:
            legend = f"{average} average AP = {precision_avg:.3f}"
            sns.lineplot(x=rec_avg, y=pre_avg, label=legend, linewidth=4, color="navy", linestyle="--")

            axes.set_xlabel("recall")
            axes.set_ylabel("precision")
            axes.set_title(f"precision vs. recall curve (Weighted AP = {precision_overall:.3f})")
            axes.set_xlim([0, 1])
            axes.set_ylim([0, 1])

            plt.legend(loc="best")
        return results

    def roc_auc_score(self, multi_class: str = "ovo", average: str = "macro") -> float:
        """

        :param multi_class: {‘ovr’, ‘ovo’}, default='ovo'
        :param average: {‘micro’, ‘macro’}, default='macro'
        """
        return roc_auc_score(self.y_actual_binary, self.y_scores, multi_class = multi_class, average = average)


    def roc_curve(self, multi_class: str = "ovo", average: str = "macro", plot: bool = True, figsize=(5, 5)) -> dict:
        """

        :param multi_class: {‘ovr’, ‘ovo’}, default='ovo'
        :param average: {‘micro’, ‘macro’}, default='macro'
        :param plot: whether to plot or not
        :param figsize:
        :return: a dictionary {label : df[threshold, fpr, tpr]}
        """
        if self.binary:
            average = 'macro'  # binary only support macro, or will be mistakenly wrong

        results = dict()
        aucs = dict()

        # fpr, tpr and AUC by class
        for i, label in enumerate(self.y_labels):
            fpr, tpr, threshold = roc_curve(self.y_actual_binary[:, i], self.y_scores[:, i], drop_intermediate=False)
            results[label] = pd.DataFrame({
                'threshold': threshold,
                'fpr': fpr,
                'tpr': tpr
            })
            aucs[label] = auc(fpr, tpr)

        # average ROC, AUC (micro/macro)
        if average == 'micro':
            ## micro ROC
            fpr_avg, tpr_avg, threshold_avg = roc_curve(self.y_actual_binary.ravel(), self.y_scores.ravel(),
                                                        drop_intermediate=False)
        elif average == 'macro':
            ## macro ROC
            all_fpr = np.unique(np.concatenate([results[label]['fpr'] for label in self.y_labels]))

            #### Then interpolate all ROC curves at this points
            mean_tpr = np.zeros_like(all_fpr)
            for label in self.y_labels:
                mean_tpr += np.interp(all_fpr, results[label]['fpr'], results[label]['tpr'])

            mean_tpr /= len(self.y_labels)

            fpr_avg = all_fpr
            tpr_avg = mean_tpr
            threshold_avg = np.empty(fpr_avg.shape)
            threshold_avg[:] = np.nan

        else:
            raise ValueError("parameter 'average' can only be micro or macro or binary")

        results['AVG'] = pd.DataFrame({
            'threshold': threshold_avg,
            'fpr': fpr_avg,
            'tpr': tpr_avg
        })

        roc_auc_avg = self.roc_auc_score( multi_class=multi_class, average=average)
        roc_auc_overall = self.roc_auc_score(multi_class=multi_class, average='weighted')  # weighted AUC

        if plot:
            sns.set_style("ticks", {'axes.grid': True})
            fig, axes = plt.subplots(1, 1, figsize=figsize)

            # ROC by class
            for i, label in enumerate(self.y_labels):
                legend = f"{label} (AUC = {aucs[label]:.3f})"
                sns.lineplot(x=results[label]['fpr'], y=results[label]['tpr'], label=legend, linewidth=2)

            # average ROC:
            legend = f"{average} AUC = {roc_auc_avg:.3f}"
            sns.lineplot(x=fpr_avg, y=tpr_avg, label=legend, linewidth=4, color="navy", linestyle="--")

            axes.set_xlabel("fpr")
            axes.set_ylabel("tpr")
            axes.set_title(f"ROC curve (Weighted AUC = {roc_auc_overall:.3f})")
            axes.set_xlim([0, 1])
            axes.set_ylim([0, 1])

            plt.legend(loc="best")
        return results


    def precision_recall_roc_by_class(self, class_, figsize=(12, 5)):
        """

        :param class_: class name string
        :return:
        """
        precision_recall = self.precision_recall(plot=False).get(class_)
        fpr_tpr = self.roc_curve(plot=False).get(class_)

        if precision_recall is None or fpr_tpr is None:
            raise ValueError(f"class_ {class_} not found in target variable, current target list = {self.y_labels}")

        sns.set_style("ticks", {'axes.grid': True})
        fig, axes = plt.subplots(1, 2, figsize=figsize)

        sns.lineplot(ax=axes[0], x=precision_recall['threshold'], y=precision_recall['precision'], label='Precision',
                     linewidth=2)
        sns.lineplot(ax=axes[0], x=precision_recall['threshold'], y=precision_recall['recall'], label='Recall',
                     linewidth=2)

        sns.lineplot(ax=axes[1], x=fpr_tpr['threshold'], y=fpr_tpr['tpr'], label='sensitivity(tpr)', linewidth=2)
        sns.lineplot(ax=axes[1], x=fpr_tpr['threshold'], y=1 - fpr_tpr['fpr'], label='specificity(1-fpr)', linewidth=2)

        axes[0].set_xlabel("threshold")
        axes[0].set_title("Precision/Recall vs. Threshold")
        axes[0].set_xlim([0, 1])
        axes[0].set_ylim([0, 1])

        axes[1].set_xlabel("threshold")
        axes[1].set_title("Sensitivity/Specificity vs. Threshold")
        axes[1].set_xlim([0, 1])
        axes[1].set_ylim([0, 1])

        plt.legend(loc="best")

    def report(self, output_dict=False, sample_weight=None):
        """

        :param output_dict: if True, will return report in dict format, False will only print
        :param sample_weight: array-like of shape (n_samples,)
        :return:
        """
        return classification_report(self.y_actual, self.y_predict_cv, target_names=self.y_labels,
                                     output_dict=output_dict, sample_weight=sample_weight)

class OrdinalClfValidator(Validator):
    def __init__(self, clf, Xs, y_actual, ordered_classes):
        self.clf = clf
        self.Xs = Xs
        self.y_actual = y_actual
        self.y_predict = clf.predict(Xs)
        self.y_labels = ordered_classes

        # label encode
        self.label_mappings = dict(zip(range(len(ordered_classes)), ordered_classes))
        self.reverse_label_mappings = dict(zip(ordered_classes, range(len(ordered_classes))))
        self.y_actual_label = pd.Series(self.y_actual).map(self.reverse_label_mappings).values
        self.y_predict_label = pd.Series(self.y_predict).map(self.reverse_label_mappings).values

    def conf_mat(self, error: bool = False) -> np.ndarray:

        cm = confusion_matrix(self.y_actual, self.y_predict)
        cm_sum = np.sum(cm, axis=1, keepdims=True)
        cm_perc = cm / cm_sum.astype(float)
        np.fill_diagonal(cm_perc, 0)  # fill diagonal value for error chart to zero

        return cm_perc if error else cm

    def rank_matching(self, rate: bool = False, plot: bool = True) -> pd.DataFrame:
        # similar to 30/60/90: accuracy for: exact match/within 1 notch/within 2 notch,etc.
        n_ = len(self.y_actual_label)
        k_ = len(self.y_labels)

        # calculate diffs
        diff = self.y_predict_label - self.y_actual_label
        abs_diff = np.abs(diff)
        # calculate absolute matching numbers
        ## ensure the number of index is the same as total possible notches, i.e., still work when small errors
        matching = pd.Series(index=np.arange(-k_ + 1, k_), dtype='int')
        matching.update(pd.Series(diff).value_counts().sort_index())
        abs_matching = pd.Series(index=np.arange(k_), dtype='int')
        abs_matching.update(pd.Series(abs_diff).value_counts().sort_index())
        abs_matching.index = [f"±{idx}" for idx in range(k_)]
        abs_matching_cum = abs_matching.cumsum()
        abs_matching_cum.index = ["Exact"] + [f"w/i ±{idx}" for idx in range(1, k_)]
        # calculate absolute matching numbers by class
        ## ensure the number of index is the same as total possible notches, i.e., still work when small errorsx
        abs_matching_matrix = pd.DataFrame(index=range(k_), columns=self.y_labels)
        abs_matching_matrix.index.name = 'abs_diff'
        abs_matching_matrix.columns.name = 'actual'
        amnbc = pd.DataFrame({'actual': self.y_actual, 'abs_diff': abs_diff}).reset_index()
        abs_matching_matrix.update(pd.pivot_table(amnbc, values='index', index='abs_diff',
                                                  columns='actual', aggfunc='count').fillna(0).astype(
            'int').sort_index())
        abs_matching_matrix.index = [f"±{idx}" for idx in range(len(abs_matching.index))]
        abs_matching_matrix = abs_matching_matrix.fillna(0).astype('int')
        abs_matching_cum_matrix = abs_matching_matrix.cumsum(axis=0)
        abs_matching_cum_matrix.index = ["Exact"] + [f"w/i ±{idx}" for idx in range(1, k_)]

        # calculate matching rates
        matching_rt = matching / n_
        abs_matching_rt = abs_matching / n_
        abs_matching_cum_rt = abs_matching_cum / n_
        abs_matching_cum_rt_matrix = abs_matching_cum_matrix / np.repeat(
            np.sum(abs_matching_matrix.values, axis=0, keepdims=True), repeats=k_, axis=0)

        if plot:
            fig, axes = plt.subplot_mosaic(
                [['notch_rt', 'notch_rt', 'notch_grid1'],
                 ['abs_notch_rt', 'abs_notch_rt', 'notch_grid2']],
                figsize=(15, 10), constrained_layout=True)

            # 1. plot diff by notches
            sat = (np.abs(matching_rt.index) + 1) / (matching_rt.index.max() + 1)
            sns.barplot(x=matching_rt.index, y=matching_rt.values, ax=axes['notch_rt'], hue=np.abs(matching_rt.index),
                        palette="flare", dodge=False)
            for i in axes['notch_rt'].containers:
                axes['notch_rt'].bar_label(i, )
            axes['notch_rt'].set_title("Errors by notch")
            axes['notch_rt'].set_ylim([0, 1])

            # 2. plot abs_diff by class and as a whole piece
            colors = cycle(sns.color_palette("Set2"))
            for col in abs_matching_cum_rt_matrix.columns:
                sns.lineplot(x=abs_matching_cum_rt_matrix.index, y=abs_matching_cum_rt_matrix[col],
                             ax=axes['abs_notch_rt'], lw=2,
                             linestyle="--", label=col, color=next(colors))
            l_all = sns.lineplot(x=abs_matching_cum_rt.index, y=abs_matching_cum_rt.values, ax=axes['abs_notch_rt'],
                                 color="#24201E",
                                 marker='o', markersize=10, lw=5, label='Total')
            for idx, rt in abs_matching_cum_rt.iteritems():
                # l_all.text(idx, rt + 0.05, f"{rt:.0%}", horizontalalignment = 'center')
                l_all.annotate(f"{rt:.0%}", xy=(idx, rt), xytext=(idx, rt), textcoords='offset points',
                               bbox=dict(boxstyle="round4,pad=.5", fc="0.9"), horizontalalignment='center')
            axes['abs_notch_rt'].set_title("Matching rate by abs notch")
            axes['abs_notch_rt'].set_ylim([0, 1.1])

            # 3. plot heatmap for abs diff grid
            sns.heatmap(data=abs_matching_cum_matrix, annot=True, fmt="d", ax=axes['notch_grid1'])
            axes['notch_grid1'].set_title("Matching grid by class")

            annots = abs_matching_cum_rt_matrix.applymap(lambda x: f"{x:.0%}").values
            sns.heatmap(data=abs_matching_cum_rt_matrix, annot=annots, fmt="", cmap=plt.cm.gray, ax=axes['notch_grid2'])
            axes['notch_grid2'].set_title("Matching rate grid by class")

        return abs_matching_cum_rt_matrix if rate else abs_matching_cum_matrix

    def kendall_tau(self, alpha: float = 0.95) -> tuple:
        k, p = stats.kendalltau(self.y_actual, self.y_predict, nan_policy='omit')
        sig_flag = (p < (1 - alpha) / 2)
        return k, p, sig_flag

class ClfValidatorCV(ClfValidator):

    def __init__(self, clf, Xs, y_actual, y_labels=None, cv=3, binary:bool = False):
        self.binary = binary
        self.clf = clf
        self.Xs = Xs
        self.y_actual = y_actual
        self.y_predict_cv = cross_val_predict(clf, Xs, y_actual, cv=cv)
        self.y_scores = cross_val_predict(clf, Xs, y_actual, cv=cv, method="predict_proba")

        self.cv = cv
        # map y labels
        if y_labels is None:
            self.y_labels = clf.classes_
        else:
            self.y_labels = y_labels

        self.y_actual_binary = label_binarize(self.y_actual, classes=self.y_labels)
        if len(self.y_labels) == 2 or binary:  # adjust for binary classifier -- increase to 2D dimension
            self.y_actual_binary = np.c_[1 - self.y_actual_binary, self.y_actual_binary]



class ModelSelector:
    pass

class ClfModelSelector(ModelSelector):
    def __init__(self, clf_dict: dict, Xs, y_actual, y_labels = None, binary:bool = False):
        self.binary = binary
        self.models = {k : ClfValidator(clf, Xs, y_actual, y_labels, binary = binary) for k, clf in clf_dict.items()}

    def getClfv(self, model_name: str):
        return self.models.get(model_name)

    def confusion_matrix(self, plot: bool = True, figsize = (12, 5)) -> dict:
        conf_mats = {model_name : clf.confusion_matrix(plot = False) for model_name, clf in self.models.items()}

        n_models = len(self.models)
        figsize = (figsize[0], figsize[1] * n_models)
        if plot:
            fig, axes = plt.subplots(n_models, 2, sharex=True, figsize=figsize)

            for idx, (model_name, (cm, cm_perc)) in enumerate(conf_mats.items()):
                # annotation
                annot1 = np.empty_like(cm).astype(str)  # absolute value chart (colorful heatmap)
                annot2 = np.empty_like(cm).astype(str)  # error chart (gray scale)
                nrows, ncols = cm.shape
                for i in range(nrows):
                    for j in range(ncols):
                        c = cm.iloc[i, j]
                        p = cm_perc.iloc[i, j]
                        if c == 0:
                            annot1[i, j] = ''
                            annot2[i, j] = ''
                        else:
                            annot1[i, j] = f'{c : .0f}'

                            if i != j:
                                annot2[i, j] = f'{p :.0%}'
                            else:
                                annot2[i, j] = ''

                sns.heatmap(ax=axes[idx, 0], data = cm, annot = annot1, fmt = '')
                sns.heatmap(ax=axes[idx, 1], data = cm_perc, annot = annot2, fmt='', cmap=plt.cm.gray, vmin=0, vmax=1)

                axes[idx, 0].set_title(f'Confusion Matrix | {model_name}')
                axes[idx, 1].set_title(f'Error Heatmap - p(predict | actual) | {model_name}')

        return conf_mats


    def roc_curve(self, multi_class: str = "ovo", average: str = "macro", plot: bool = True, figsize=(5, 5)) -> dict:
        if self.binary:
            average = 'macro'

        avg_rocs = {}
        avg_aucs = {}
        for model_name, clfv in self.models.items():
            avg_rocs[model_name] = clfv.roc_curve(multi_class = multi_class, average = average, plot = False).get('AVG')
            avg_aucs[model_name] = clfv.roc_auc_score(multi_class = multi_class, average = average)

        if plot:
            sns.set_style("ticks", {'axes.grid': True})
            fig, axes = plt.subplots(1, 1, figsize=figsize)

            # ROC by model
            for model_name, avg_roc in avg_rocs.items():
                legend = f"{model_name} ({average} AUC = {avg_aucs[model_name]:.3f})"
                sns.lineplot(x = avg_roc['fpr'], y = avg_roc['tpr'], label = legend, linewidth = 2)

            axes.set_xlabel("fpr")
            axes.set_ylabel("tpr")
            axes.set_title(f"ROC curve")
            axes.set_xlim([0, 1])
            axes.set_ylim([0, 1])
            plt.legend(loc="best")

        return avg_rocs

    def precision_recall(self, average: str = "macro", plot: bool = True, figsize = (5, 5)) -> dict:
        avg_pre_recs = {}
        avg_aps = {}
        for model_name, clfv in self.models.items():
            avg_pre_recs[model_name] = clfv.precision_recall(average = average, plot = False).get('AVG')
            avg_aps[model_name] = clfv.average_precision_score(average = average)


        if plot:
            sns.set_style("ticks", {'axes.grid': True})
            fig, axes = plt.subplots(1, 1, figsize=figsize)

            # ROC by model
            for model_name, avg_pre_rec in avg_pre_recs.items():
                legend = f"{model_name} ({average} AP = {avg_aps[model_name]:.3f})"
                sns.lineplot(x = avg_pre_rec['precision'], y = avg_pre_rec['recall'], label = legend, linewidth = 2)

            axes.set_xlabel("precision")
            axes.set_ylabel("recall")
            axes.set_title(f"precision vs. recall curve")
            axes.set_xlim([0, 1])
            axes.set_ylim([0, 1])
            plt.legend(loc="best")

        return avg_pre_recs

class ClfModelSelectorCV(ClfModelSelector):
    def __init__(self, clf_dict: dict, Xs, y_actual, y_labels = None, binary:bool = False):
        self.binary = self.binary
        self.models = {k : ClfValidatorCV(clf, Xs, y_actual, y_labels, binary = binary) for k, clf in clf_dict.items()}

class OrdinalClfModelSelector(ModelSelector):
    def __init__(self, clf_dict: dict, Xs, y_actual, ordered_classes):
        self.models = {k : OrdinalClfValidator(clf, Xs, y_actual, ordered_classes) for k, clf in clf_dict.items()}

    def rank_matching(self, plot: bool = True, by_class: str = None) -> pd.DataFrame:
        matching_mats = pd.DataFrame(columns = ['model_name', 'band', 'matching'])
        for model_name, oclfv in self.models.items():
            if by_class:
                s = oclfv.rank_matching(rate = True, plot = False)[by_class]
                matching_mats = matching_mats.append(pd.DataFrame({'model_name' : model_name, 'band' : s.index, 'matching' : s.values}), ignore_index = True)
            else:
                s = oclfv.rank_matching(rate = False, plot = False).sum(axis = 1)
                s = s / s[-1]
                matching_mats = matching_mats.append(pd.DataFrame({'model_name' : model_name, 'band' : s.index, 'matching' : s.values}), ignore_index = True)

        if plot:
            fig, axes = plt.subplots(1, 1, figsize = (8, 8))

            sns.lineplot(data = matching_mats, x = 'band', y = 'matching', hue = 'model_name', lw = 3, marker="o", markersize = 10, dashes=False)
            if by_class:
                axes.set_title(f'Matching rate by notch - class {by_class}')
            else:
                axes.set_title('Matching rate by notch')

        return matching_mats

    def kendall_tau(self, alpha: float = 0.95, plot: bool = True) -> pd.DataFrame:
        kts = []
        for model_name, oclfv in self.models.items():
            kt, p, sig_flag = oclfv.kendall_tau(alpha = alpha)
            kts.append({'model_name' : model_name, 'kt' : kt, 'p' : p, 'sig' : sig_flag})

        kts = pd.DataFrame.from_records(kts)

        if plot:
            fig, axes = plt.subplots(1, 1, figsize = (6, 6))

            sns.barplot(data = kts, x = 'model_name', y = 'kt', palette = 'Set2')
            for idx, r in kts.iterrows():
                kt = r['kt']
                sig = r['sig']
                fc = '#90EC48' if sig else '#F5765A'  #if significant the facecolor of label will be green, otherwise red
                axes.annotate(f"{kt:.1%}", xy=(idx, kt), xytext=(idx, kt), textcoords='offset points',
                    bbox = dict(boxstyle="round4,pad=.5", fc = fc, alpha=0.8), horizontalalignment = 'center')

            axes.set_ylim([0, 1])
            axes.set_title('Kendall Tau correlation')

        return kts


class OrdinalClfModelSelectorCV(ClfModelSelectorCV):
    def __init__(self, clf_dict: dict, Xs, y_actual, y_labels = None):
        self.models = {k : ClfModelSelectorCV(clf, Xs, y_actual, y_labels) for k, clf in clf_dict.items()}


from typing import List, Literal, Tuple
import ibis
from ibis import _
import numpy as np
import pandas as pd
from luntaiDs.ModelingTools.Explore.engines.base import _BaseEDAEngine, _BaseNumericHelper, serialize
from luntaiDs.ModelingTools.Explore.summary import CategUniVarClfTargetCorr, DescStat, NumericUniVarClfTargetCorr, QuantileStat, StatVar, XtremeStat, \
    BinaryStatAttr, BinaryStatSummary, CategStatAttr, \
    CategStatSummary, NumericStatAttr, NumericStatSummary
        

class NumericHelperIbis(_BaseNumericHelper):
    def __init__(self, df: ibis.expr.types.Table, colname: str):
        self._df = df
        self._colname = colname
    
    def get_descriptive_stat(self) -> DescStat:
            # get min/max
        descrp = (
            self._df
            .mutate(_[self._colname].cast('Float64').name(self._colname))
            .mutate(NORM_ = (_[self._colname] - _[self._colname].mean()))
            .mutate(
                NORM3_ = _['NORM_'] * _['NORM_'] * _['NORM_'],
                NORM4_ = _['NORM_'] * _['NORM_'] * _['NORM_'] * _['NORM_'],
                NORMABS_ = _['NORM_'].abs()
            )
            .aggregate(
                MIN_ = _[self._colname].min(),
                MAX_ = _[self._colname].max(),
                MEAN_ = _[self._colname].mean(),
                VAR_ = _[self._colname].var(),
                STD_ = _[self._colname].std(),
                MAD_ = _['NORMABS_'].mean(),
                NORM3_ = _['NORM3_'].mean(),
                NORM4_ = _['NORM4_'].mean()
            )
            .to_pandas()
            .astype('float')
        )
        
        mean_ = descrp.loc[0, 'MEAN_']
        var_ = descrp.loc[0, 'VAR_']
        std_ = descrp.loc[0, 'STD_']
        mad_ = descrp.loc[0, 'MAD_'] # avg(abs(X-u))
        norm3_ = descrp.loc[0, 'NORM3_'] # avg((X-u)^3)
        norm4_ = descrp.loc[0, 'NORM4_'] # avg((X-u)^4)
        
            
        # descriptive statistics
        stat_descriptive = DescStat(
            mean=serialize(mean_),
            var=serialize(var_),
            std=serialize(std_),
            skew=serialize(norm3_ / std_ ** 3),
            kurt=serialize(norm4_ / std_ ** 4) - 3, # normal is 3
            mad=serialize(mad_),
            cv=serialize(std_ / mean_),
            normality_p=None,
        )
        return stat_descriptive
    
    def getQuantiles(self, quantiles: List[float]) -> List[float]:
        kws = dict()
        for q in quantiles:
            kws[str(q)] = _[self._colname].quantile(q)
        qs = self._df.aggregate(**kws).to_pandas()
        return qs.iloc[0, :].astype('float').tolist()
    
    def get_quantile_stat(self) -> QuantileStat:
        descrp = (
            self._df
            .aggregate(
                MIN_ = _[self._colname].min(),
                MAX_ = _[self._colname].max(),
            )
            .to_pandas()
            .astype('float')
        )
        
        min_ = descrp.loc[0, 'MIN_']
        max_ = descrp.loc[0, 'MAX_']
        
        # get quantiles
        try:
            q001, q005, q025, q05, q075, q095, q099 = self.getQuantiles(
                self._df, 
                colname=self._colname, 
                quantiles=[0.01, 0.05, 0.25, 0.5, 0.75, 0.95, 0.99]
            )
            
        except Exception as e:
            # use ntile as backup
            quantiles = (
                self._df
                .mutate(
                    NTILE_ = ibis.ntile(100).over(
                        order_by=_[self._colname]
                    )
                )
                .group_by('NTILE_')
                .aggregate(
                    MIN_ = _[self._colname].min(), # 0=0%, 1=1%, 2=2%
                    #MAX_ = _[colname].max(),
                )
                .to_pandas()
                .set_index('NTILE_')
                .astype('float')
            )
            q001 = quantiles.loc[1, 'MIN_']
            q005 = quantiles.loc[5, 'MIN_']
            q025 = quantiles.loc[25, 'MIN_']
            q05 = quantiles.loc[50, 'MIN_']
            q075 = quantiles.loc[75, 'MIN_']
            q095 = quantiles.loc[95, 'MIN_']
            q099 = quantiles.loc[99, 'MIN_']
            
            
        stat_quantile = QuantileStat(
            minimum=serialize(min_),
            perc_1th=serialize(q001),
            perc_5th=serialize(q005),
            q1=serialize(q025),
            median=serialize(q05),
            q3=serialize(q075),
            perc_95th=serialize(q095),
            perc_99th=serialize(q099),
            maximum=serialize(max_),
            iqr=serialize(q075 - q025),
            range=serialize(max_ - min_),
        )

        return stat_quantile        
        
    
    def get_xtreme_stat(self, xtreme_method: Literal["iqr", "quantile"] = "iqr") -> XtremeStat:
        if xtreme_method == "iqr":
            try:
                q1, q3 = self.getQuantiles(
                    self._df, 
                    colname=self._colname, 
                    quantiles=[0.25, 0.75]
                )
            except Exception as e:
                # use ntile as backup
                quantiles = (
                    self._df
                    .mutate(
                        NTILE_ = ibis.ntile(4).over(
                            order_by=_[self._colname]
                        )
                    )
                    .group_by('NTILE_')
                    .aggregate(
                        MIN_ = _[self._colname].min(), # 0=0%, 1=25%, 2=50%
                        #MAX_ = _[colname].max(),
                    )
                    .to_pandas()
                    .set_index('NTILE_')
                    .astype('float')
                )
                q1 = quantiles.loc[1, 'MIN_']
                q3 = quantiles.loc[3, 'MIN_']
            
            iqr = q3 - q1
            lbound = q1 - 1.5 * iqr
            rbound = q3 + 1.5 * iqr

        elif xtreme_method == "quantile":
            try:
                lbound, rbound = self.getQuantiles(
                    self._df, 
                    colname=self._colname, 
                    quantiles=[0.01, 0.99]
                )
                
            except Exception as e:
                # use ntile as backup
                quantiles = (
                    self._df
                    .mutate(
                        NTILE_ = ibis.ntile(100).over(
                            order_by=_[self._colname]
                        )
                    )
                    .group_by('NTILE_')
                    .aggregate(
                        MIN_ = _[self._colname].min(), # 0=0%, 1=1%, 2=2%
                        #MAX_ = _[self._colname].max(),
                    )
                    .to_pandas()
                    .set_index('NTILE_')
                    .astype('float')
                )
                lbound = quantiles.loc[1, 'MIN_']
                rbound = quantiles.loc[99, 'MIN_']
        else:
            raise ValueError("xtreme_method can only be iqr or quantile")

        xtremes = (
            self._df
            .aggregate(
                lxtreme_mean = _[self._colname].mean(
                    where = _[self._colname] < lbound
                ),
                lxtreme_median = _[self._colname].approx_median(
                    where = _[self._colname] < lbound
                ),
                rxtreme_mean = _[self._colname].mean(
                    where = _[self._colname] > rbound
                ),
                rxtreme_median = _[self._colname].approx_median(
                    where = _[self._colname] > rbound
                )
            )
            .to_pandas()
            .astype('float')
        )

        return XtremeStat(
            lbound=serialize(lbound),
            rbound=serialize(rbound),
            lxtreme_mean=serialize(xtremes.loc[0, 'lxtreme_mean']),
            lxtreme_median=serialize(xtremes.loc[0, 'lxtreme_median']),
            rxtreme_mean=serialize(xtremes.loc[0, 'rxtreme_mean']),
            rxtreme_median=serialize(xtremes.loc[0, 'rxtreme_median']),
        )
        
    def get_histogram(self, n_bins:int) -> Tuple[np.ndarray, np.ndarray]:
        # get bin edges
        stat = (
            self._df
            .aggregate(
                MIN_ = _[self._colname].min(),
                MAX_ = _[self._colname].max()
            )
            .to_pandas()
            .astype('float')
        )
        min_ = stat.loc[0, 'MIN_']
        max_ = stat.loc[0, 'MAX_']
        bin_edges_ = np.linspace(min_, max_, n_bins + 1)
        
        # get count of each bins
        hist = (
            self._df
            .mutate(
                _[self._colname]
                .histogram(nbins = n_bins)
                # need to clip because the max value will be at edge
                .clip(lower = 0, upper = n_bins - 1)
                .name('HIST_')
            )
            .group_by('HIST_')
            .aggregate(
                # MIN_ = _[self._colname].min(),
                # MAX_ = _[self._colname].max(),
                NUM_ = _[self._colname].count()
            )
            .to_pandas()
            .dropna()
            .set_index('HIST_')
        )
        hist_ = (
            pd.DataFrame(
                index = np.arange(n_bins)
            )
            .join(
                hist,
                how = 'left'
            )
            .fill_null(0)
            ['NUM_'].astype('int')
            .values
        )
        return hist_, bin_edges_

def _combine_x_y_ibis(df: ibis.expr.types.Table, 
        combine_x_categ: bool = False) -> ibis.expr.types.Table:
    df = (
        df
        .dropna(
            subset = ['y'], 
            how = 'any'
        )
    )
    # categorical variable for x only:
    if combine_x_categ:
        vc = df['x'].topk(20)['x'].to_pandas() # only keep top 20 categories
        df = (
            df
            .mutate(
                _['x']
                .isin(vc.tolist())
                .ifelse(_['x'], 'Others')
                .name('x')
            )
        )
    return df.select('x', 'y')


class EDAEngineIbis(_BaseEDAEngine):
    def __init__(self, df: ibis.expr.types.Table):
        self._df = df
        
    def get_columns(self) -> List[str]:
        """get all column list from given dataset

        :return List[str]: list of columns in the dataset
        """
        return self._df.columns
    
    def _fit_common_categ(self, colname: str, attr: CategStatAttr) -> CategStatSummary:
        """common categorical variable fitting can be reused by subclass categorical fitting

        :param str colname: column name
        :param CategStatAttr attr: categorical variable attribute object
        :return CategStatSummary: categorical variable summary object
        """
        basic = (
            self._df
            .aggregate(
                total_ = _.count(),
                missing_ = _.count(where = _[colname].isnull()),
                unique_ = _[colname].nunique() # already dropped na
            )
            .to_pandas()
        )
        # total number of records
        total_ = serialize(basic.loc[0, 'total_'])
        # missing rate
        missing = basic.loc[0, 'missing_']
        missing_ = StatVar(
            value=serialize(missing), 
            perc=serialize(missing / total_)
        )
        # num of unique value (dropped NA value)
        unique = basic.loc[0, 'unique_']
        unique_ = StatVar(
            value=serialize(unique), 
            perc=serialize(unique / total_)
        )
        
        distribution = (
            self._df
            .dropna([colname])
            [colname]
            .value_counts()
            .order_by(ibis.desc(f"{colname}_count"))
            .to_pandas()
        )

        # category counts without null
        vcounts_ = pd.Series(
            distribution.iloc[:, 1].values,
            index = distribution.iloc[:, 0],
            name = colname
        )
        vpercs_ = (
            vcounts_ / vcounts_.sum()
        ).fill_null(0)
        
        return CategStatSummary(
            colname_ = colname,
            total_ = total_,
            missing_ = missing_,
            unique_ = unique_,
            vcounts_ = vcounts_,
            vpercs_ = vpercs_
        )
        
    def fit_binary(self, colname: str, attr: BinaryStatAttr) -> BinaryStatSummary:
        """binary categorical variable fitting

        :param str colname: column name
        :param BinaryStatAttr attr: binary variable attribute object
        :return BinaryStatSummary: binary variable summary object
        """
        # reuse common stat
        common_stat_summary = self._fit_common_categ(colname, attr)
        # binary stat
        binary = (
            ibis
            .case()
            .when(self._df[colname].isnull(), int(attr.na_to_pos_))
            .when(self._df[colname].isin(attr.pos_values_), 1)
            .else_(0)
            .end()
            .value_counts()
            .to_pandas()
        )
        binary_vcounts_ = pd.Series(
            binary.iloc[:, 1].values,
            index = binary.iloc[:, 0],
            name = colname
        )
        binary_vpercs_ = (
            binary_vcounts_ / binary_vcounts_.sum()
        ).fill_null(0)
        
        return BinaryStatSummary(
            colname_ = common_stat_summary.colname_,
            total_ = common_stat_summary.total_,
            missing_ = common_stat_summary.missing_,
            unique_ = common_stat_summary.unique_,
            vcounts_ = common_stat_summary.vcounts_,
            vpercs_ = common_stat_summary.vpercs_,
            binary_vcounts_ = binary_vcounts_,
            binary_vpercs_ = binary_vpercs_
        )
        
    def fit_numeric(self, colname: str, attr: NumericStatAttr) -> NumericStatSummary:
        """numeric variable fitting

        :param str colname: column name
        :param NumericStatAttr attr: numeric variable attribute object
        :return NumericStatSummary: numeric variable summary object
        """
        df = self._df.mutate(_[colname].cast('Float64').name(colname))
        basic = (
            df
            .aggregate(
                total_ = _.count(),
                missing_ = _.count(where = (_[colname].isnull()) | (_[colname].isnan())),
                zeros_ =  _.count(where = _[colname] == 0),
                infs_pos_ = _.count(where = (_[colname] > 0) & (_[colname].isinf())),
                infs_neg_ = _.count(where = (_[colname] < 0) & (_[colname].isinf())),
            )
            .to_pandas()
        )
        # total number of records
        total_ = serialize(basic.loc[0, 'total_'])
        # missing rate
        missing = basic.loc[0, 'missing_']
        missing_ = StatVar(
            value=serialize(missing), 
            perc=serialize(missing / total_)
        )
        # zero rate
        zeros = basic.loc[0, 'zeros_']
        zeros_ = StatVar(
            value=serialize(zeros), 
            perc=serialize(zeros / total_)
        )
        # infinite value
        infs_pos = basic.loc[0, 'infs_pos_']
        infs_pos_ = StatVar(
            value=serialize(infs_pos), 
            perc=serialize(infs_pos / total_)
        )
        infs_neg = basic.loc[0, 'infs_neg_']
        infs_neg_ = StatVar(
            value=serialize(infs_neg), 
            perc=serialize(infs_neg / total_)
        )
        
        # clean version
        df = (
            df
            .dropna([colname])
            .filter(~_[colname].isinf())
        )
        if attr.setaside_zero_:
            df = df.filter(_[colname] != 0)
        n_exextreme = df.count().to_pandas()
        
        # xtreme value
        if attr.xtreme_method_:
            xstat = NumericHelperIbis(
                df = df,
                colname = colname
            )
            xtreme_stat_ = xstat.get_xtreme_stat(
                xtreme_method=attr.xtreme_method_
            )
            df_clean = df.filter(
                (_[colname] >= xtreme_stat_.lbound) 
                & (_[colname] <= xtreme_stat_.rbound)
            )
            num_xtreme = (n_exextreme - df_clean.count()).to_pandas()
        else:
            df_clean = df
            num_xtreme = 0
            xtreme_stat_ = None
        
        xtreme_ = StatVar(
            value=serialize(num_xtreme), 
            perc=serialize(num_xtreme / total_)
        )
        
        # percentile & descriptive statistics
        xstat_clean = NumericHelperIbis(
            df = df_clean,
            colname = colname
        )
        stat_quantile_ = xstat_clean.get_quantile_stat()
        stat_descriptive_ = xstat_clean.get_descriptive_stat()
        
        # histogram
        hist_, bin_edges_ = xstat_clean.get_histogram(
            n_bins = attr.bins_
        )
        
        num_stat = NumericStatSummary(
            colname_ = colname,
            total_ = total_,
            missing_ = missing_,
            zeros_ = zeros_,
            infs_pos_ = infs_pos_,
            infs_neg_ = infs_neg_,
            stat_quantile_ = stat_quantile_,
            stat_descriptive_ = stat_descriptive_,
            hist_ = hist_,
            bin_edges_ = bin_edges_,
            xtreme_ = xtreme_,
            xtreme_stat_ = xtreme_stat_
        )
         
        if attr.log_scale_:
            colname_log = f"{colname}_log"
            df_log = (
                df_clean
                .mutate(
                    (_[colname] > 0)
                    .ifelse(
                        (_[colname] + 1).log10(),
                        -(1 - _[colname]).log10()
                    )
                    .name(colname_log)
                )
            )
            
            # xtreme value
            if attr.xtreme_method_:
                xstat_log = NumericHelperIbis(
                    df = df_log,
                    colname = colname_log
                )
                num_stat.xtreme_stat_log_ = xstat_log.get_xtreme_stat(
                    xtreme_method=attr.xtreme_method_
                )
                df_clean_log = df_log.filter(
                    (_[colname_log] >= num_stat.xtreme_stat_log_.lbound) 
                    & (_[colname_log] <= num_stat.xtreme_stat_log_.rbound)
                )
                num_xtreme_log = (n_exextreme - df_clean_log.count()).to_pandas()
            else:
                df_clean_log = df_log
                num_xtreme_log = 0
                num_stat.xtreme_stat_log_ = None
            
            num_stat.xtreme_log_ = StatVar(
                value=serialize(num_xtreme_log), 
                perc=serialize(num_xtreme_log / total_)
            )
            
            # percentile & descriptive statistics
            xstat_clean_log = NumericHelperIbis(
                df = df_clean_log,
                colname = colname
            )
            num_stat.stat_quantile_log_ = xstat_clean_log.get_quantile_stat()
            num_stat.stat_descriptive_log_ = xstat_clean_log.get_descriptive_stat()

            # histogram
            num_stat.hist_log_, num_stat.bin_edges_log_ = xstat_clean_log.get_histogram(
                n_bins = attr.bins_
            )
            
        return num_stat
    
    
    def fit_univarclf_categ(self, x_col: str, y_col: str) -> CategUniVarClfTargetCorr:
        """fit univariate correlation for classififer target and categorical variable

        :param str x_col: column of a categorical variable
        :param str y_col: column of the classifier target varaible
        :return CategUniVarClfTargetCorr: result object
        """
        df = (
            self._df
            .mutate(
                x = _[x_col].cast('string').fill_null('Missing_'),
                y = _[y_col].cast('string'),
            )
        )
        df = _combine_x_y_ibis(
            df,
            combine_x_categ = True # combine x's categories to be less than 20
        )
        
        # count x/y joint distribution
        num_x_y = (
            df
            .group_by(['y', 'x'])
            .aggregate(_.count().name('NUM_'))
            .to_pandas()
        )
        # categorical:  p(x | y)
        p_x_y_ = {}
        for y in num_x_y['y'].unique():
            cs = num_x_y.set_index('y').loc[y, :]
            d = pd.DataFrame({
                    'count' : cs['NUM_'].values, 
                    'perc' : cs['NUM_'].values  / cs['NUM_'].values.sum()
                }, 
                index = cs['x']
            )
            p_x_y_[y] = d.fill_null(0)
            
        # categorical:  p(y | x)  prob (event rate when binary clf) by category
        p_y_x_ = {}
        for x in num_x_y['x'].unique():
            cs = num_x_y.set_index('x').loc[x, :]
            d = pd.DataFrame({
                    'count' : cs['NUM_'].values, 
                    'perc' : cs['NUM_'].values  / cs['NUM_'].values.sum()
                }, 
                index = cs['y']
            )
            p_y_x_[x] = d.fill_null(0) #.to_dict(orient = 'index')
            
        return CategUniVarClfTargetCorr(
            colname_ = x_col,
            yname_ = y_col,
            ylabels_ = (
                self._df
                .filter(~_[y_col].isnull())
                .select(y_col)
                .distinct()
                [y_col]
                .to_pandas()
                .astype('str')
                .tolist()
            ),
            p_x_y_ = p_x_y_,
            p_y_x_ = p_y_x_
        )
    
    def fit_univarclf_numeric(self, x_col: str, y_col: str) -> NumericUniVarClfTargetCorr:
        """fit univariate correlation for classififer target and numeric variable

        :param str x_col: column of a numeric variable
        :param str y_col: column of the classifier target varaible
        :return NumericUniVarClfTargetCorr: result object
        """
        df = (
            self._df
            .mutate(
                x = _[x_col].cast('float'),
                y = _[y_col].cast('string'),
            )
        )
        df = _combine_x_y_ibis(
            df,
            combine_x_categ = False # do not combine for numeric
        )
        
        # numerical:  p(x | y)  distribution by target (boxplot)
        t = {'origin' : {}, 'log' : {}}
        y_uniques = (
            df
            .select('y')
            .distinct()
            ['y']
            .to_pandas()
            .values
        )
        for y in y_uniques:
            # EDA on given subset
            eda = EDAEngineIbis(df.filter(_['y'] == y))
            summary = eda.fit_numeric(
                colname = 'x', # already renamed
                attr = NumericStatAttr(
                    setaside_zero_ = False, 
                    log_scale_ = True, 
                    xtreme_method_ = 'iqr', 
                    bins_ = 100
                )
            )
            
            t['origin'][y] = {
                'lbound': summary.xtreme_stat_.lbound,
                'q1' : summary.stat_quantile_.q1,
                'mean' : summary.stat_descriptive_.mean,
                'median' : summary.stat_quantile_.median,
                'q3' : summary.stat_quantile_.q3,
                'rbound': summary.xtreme_stat_.rbound,
            }
            t['log'][y] = {
                'lbound': summary.xtreme_stat_log_.lbound,
                'q1' : summary.stat_quantile_log_.q1,
                'mean' : summary.stat_descriptive_log_.mean,
                'median' : summary.stat_quantile_log_.median,
                'q3' : summary.stat_quantile_log_.q3,
                'rbound': summary.xtreme_stat_log_.rbound,
            }
            
        p_x_y_ = {
            'origin' : pd.DataFrame.from_dict(t['origin'], orient='index'),
            'log' : pd.DataFrame.from_dict(t['log'], orient='index')
        }
        
        ## numerical:  p(y | x)  prob (event rate if binary clf) by bucketized x
        n_bins = 8
        pivot = (
            df
            .mutate(
                _['x']
                .isnull()
                .ifelse(
                    'Missing',
                    (
                        _['x']
                        .histogram(nbins = n_bins)
                        .clip(lower = 0, upper = n_bins - 1)
                        .cast('string')
                    )
                )
                .name('BUCKET_')
            )
            .fill_null({'x' : 0}) # no use
            .pivot_wider(
                names_from="y",
                names_prefix="Event",
                values_from="x",
                values_agg=_.count(),
                values_fill=0
            )
            .to_pandas()
        )
        # get bin edges
        stat = (
            df
            .aggregate(
                MIN_ = _['x'].min(),
                MAX_ = _['x'].max()
            )
            .to_pandas()
            .astype('float')
        )
        min_ = stat.loc[0, 'MIN_']
        max_ = stat.loc[0, 'MAX_']
        bin_edges_ = np.linspace(min_, max_, n_bins + 1)
        bins = pd.DataFrame({
            'left' : bin_edges_[:-1],
            'right' : bin_edges_[1:],
            'BUCKET_' : np.arange(n_bins).astype('str')
        })
        bins['Bin'] = (
            "[" 
            + bins['left'].round(4).astype('str') 
            + ", " 
            + bins['right'].round(4).astype('str') 
            + ")"
        )

        bin_pivot = bins.merge(
            pivot,
            on = 'BUCKET_',
            how = 'outer'
        )
        bin_pivot['Bin'] = bin_pivot['Bin'].fill_null('Missing')
        # add counts
        event_cols = (
            pivot
            .columns
            .difference(['BUCKET_']) # remove BUCKET_ column
            .tolist()
        )
        # add count
        bin_pivot['Count'] = bin_pivot[event_cols].sum(axis = 1)
        bin_pivot['Count (%)'] = bin_pivot['Count'] / bin_pivot['Count'].sum()
        # add event ratios
        ratio_cols = []
        for event_col in event_cols:
            event = event_col.replace('Event_', '')
            ratio_col = f"Event_rate_{event}"
            bin_pivot[ratio_col] = bin_pivot[event_col] / bin_pivot['Count']
            ratio_cols.append(ratio_col)

        final_cols = ['Bin', 'Count', 'Count (%)'] + event_cols + ratio_cols
        p_y_x_ = bin_pivot[final_cols]
        
        return NumericUniVarClfTargetCorr(
            colname_ = x_col,
            yname_ = y_col,
            ylabels_ = (
                self._df
                .filter(~_[y_col].isnull())
                .select(y_col)
                .distinct()
                [y_col]
                .to_pandas()
                .astype('str')
                .tolist()
            ),
            p_x_y_ = p_x_y_,
            p_y_x_ = p_y_x_
        )


from __future__ import annotations
from typing import List
from pathlib import Path
import pandas as pd
import fsspec
import logging
import ibis
from datetime import date
from luntaiDs.CommonTools.SnapStructure.tools import match_date_from_str
from luntaiDs.CommonTools.dtyper import DSchema
from luntaiDs.CommonTools.utils import save_ibis_to_parquet_on_fs, str2dt

    
class SnapshotDataManagerBase:
    @classmethod
    def setup(cls,):
        pass

    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema , e.g., RAW, PROCESSED
        :param table:  the virtual table under each schema, e.g., CLNT_GENERAL, BDA_GENERAL
        """
        self.schema = schema
        self.table = table

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(schema={self.schema}; table={self.table})"
    
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        raise NotImplementedError("")
    
    def init_table(self, *args, **kws):
        """initialize schema/table

        """
        raise NotImplementedError("")
    
    def get_schema(self) -> ibis.expr.schema.Schema:
        """Check dtypes of the given schema/dataset

        :return: the ibis schema of the given table
        """
        return self.read_random_snap().schema()

    
    def pre_save_check(self, snap_dt: date, overwrite: bool = False) -> bool:
        """pre-check before inserting records, will delete existing records at snap_dt if overwrite is set to True

        :param snap_dt:
        :param overwrite: if False, will not insert record if exists; will delete otherwise
        :return: True = clear to go, False = exit in downstream
        """
        if overwrite:
            self.delete(snap_dt = snap_dt)
            return True
        else:
            num_records = self.count(snap_dt)
            if num_records > 0:
                logging.warning(f"Presave check failed, {num_records} existing records found for "
                                f"{self.schema}.{self.table}@{snap_dt} while overwrite is set to False")
                return False
            return True
    
    def _save(self, df: ibis.expr.types.Table, snap_dt: date, **kws):
        """The pure logic to save ibis dataframe to the system, without handling existing record problem

        :param ibis.expr.types.Table df: the ibis dataframe
        :param date snap_dt: the snap date to save
        """
        raise NotImplementedError("")
    
    def save(self, df: ibis.expr.types.Table, snap_dt: date, overwrite: bool = False, **kws):
        """save pandas dataframe to the system

        :param ibis.expr.types.Table df: the ibis dataframe
        :param date snap_dt: the snap date to save
        :param overwrite snap_dt: whether to overwrite the existing record if any
        :raises NotImplementedError: _description_
        """
        if not self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite):
            return
        
        logging.info(f"Saving dataframe, schema {df.schema()} into {self.schema}.{self.table}")
        self._save(df = df, snap_dt = snap_dt, **kws)
        logging.info(f"Successfully saved {df.count()} records to {self.schema}.{self.table}@{snap_dt}")
        
    def get_existing_snap_dts(self) -> List[date]:
        """get existing snap dates as a list

        :return List[date]: list of snap dates
        """
        raise NotImplementedError("")
    
    def count(self, snap_dt: date) -> int:
        """count the sample size for given snap date

        :param date snap_dt: _description_
        :return int: # of rows of the dataframe
        """
        if snap_dt not in self.get_existing_snap_dts():
            return 0
        df = self.read(snap_dt=snap_dt)
        return df.count().to_pandas() # convert to scalar
    
    def read(self, snap_dt: date, **kws) -> ibis.expr.types.Table:
        """Read as ibis dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        raise NotImplementedError("")
    
    def reads(self, snap_dts: List[date], **kws) -> ibis.expr.types.Table:
        """reads as ibis dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        raise NotImplementedError("")

    def read_random_snap(self) -> ibis.expr.types.Table:
        """Select a random snap from the schema.table

        :return:
        """
        from random import choice

        existing_snaps = self.get_existing_snap_dts()
        random_snap = choice(existing_snaps)
        return self.read(random_snap)
    
    def read_pd(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        raise NotImplementedError("")
    
    def reads_pd(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        raise NotImplementedError("")
    
    def read_random_pd_snap(self) -> pd.DataFrame:
        """Select a random snap from the schema.table

        :return:
        """
        from random import choice

        existing_snaps = self.get_existing_snap_dts()
        random_snap = choice(existing_snaps)
        return self.read_pd(random_snap)
    
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        raise NotImplementedError("")

    def drop(self):
        """drop the whole table

        :return:
        """
        raise NotImplementedError("")

    def migrate(self, dst_schema: str, dst_table: str):
        """move from the existing schema.table to new one, the existing one will be deleted

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = self.duplicate(dst_schema, dst_table)
        self.drop()
        return new

    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerBase:
        """duplicate the existing schema.table to new one, the existing one will be kept

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        raise NotImplementedError("")

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date on disk

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        raise NotImplementedError("")
    
    def is_valid(self, snap_dt, rows_threshold: int = 100) -> bool:
        rows = self.count(snap_dt = snap_dt)
        return rows > rows_threshold
    
    
class SnapshotDataManagerFileSystem(SnapshotDataManagerBase):
    """files are saved as parquet snapshots under each schema.table
        file naming convention: dir/tablename_YYYY-MM-DD.parquet
        using fsspec engine for different implementation of FS
            - local
            - s3/minio
            - gcs
            - azure blob storage
            - http FS
    """
    @classmethod
    def setup(cls, fs: fsspec.AbstractFileSystem, root_dir: str):
        super(SnapshotDataManagerFileSystem, cls).setup()
        cls.ROOT_DIR = root_dir
        #cls._fs: fsspec.AbstractFileSystem = ArrowFSWrapper(fs)
        cls._fs = fs
        # duckdb engine to read table
        cls._ibis_con = ibis.duckdb.connect()
        cls._ibis_con.register_filesystem(fs)
        
    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema
        :param table:  the virtual table under each schema
        """
        super().__init__(schema = schema, table = table)
        dir = Path(self.ROOT_DIR) / schema / table
        self.dir: str = dir.as_posix()  # the root path of the table under each schema
        self.init_table()
        
    def init_table(self):
        """initialize schema/table
        
        https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.archive.AbstractArchiveFileSystem.info
        """
        self._fs.makedirs(
            path = self.dir,
            exist_ok = True
        )
        
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        return self._fs.exists(path = self.dir)
    
    def get_filename(self, snap_dt: date) -> str:
        """get parquet file name

        :param date snap_dt: the snap date of the file
        :return str: the file name ending with .parquet, excluding filepath
        """
        return f"{self.table}_{snap_dt}.parquet"
    
    def get_file_path(self, snap_dt: date) -> str:
        """the absolute path to the parquet file

        :param date snap_dt: the snap date of the file
        :return str: absolute file path
        """
        return (Path(self.dir) / self.get_filename(snap_dt)).as_posix()
    
    def get_file_path_with_protocol(self, snap_dt: date) -> str:
        """the absolute path to the parquet file, including FS protocol

        :param date snap_dt: the snap date of the file
        :return str: e.g., in s3fs, return s3://bucket/file/path/to/your.parquet
        """
        return self._fs.unstrip_protocol(
            name = self.get_file_path(snap_dt)
        )
    
    def _save(self, df: ibis.expr.types.Table, snap_dt: date, **kws):
        """The pure logic to save ibis dataframe to the system, without handling existing record problem

        :param ibis.expr.types.Table df: the ibis dataframe
        :param date snap_dt: the snap date to save
        """
        save_ibis_to_parquet_on_fs(
            df = df,
            fs = self._fs,
            filepath = self.get_file_path(snap_dt)
        )
                    
    def get_existing_snap_dts(self) -> List[date]:
        """get existing snap dates as a list

        :return List[date]: list of snap dates
        """
        existing_files: List[str] = self._fs.ls(
            path = self.dir, 
            detail = False
        )
        filenames = [Path(file).name for file in existing_files]
        existing_snaps = list(set(str2dt(match_date_from_str(f)) for f in filenames))
        existing_snaps.sort()
        return existing_snaps
                        
    def read(self, snap_dt: date, **kws) -> ibis.expr.types.Table:
        """Read as ibis dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_file_path_with_protocol(snap_dt=snap_dt)
        
        return self._ibis_con.read_parquet(filepath, **kws)
        
    
    def reads(self, snap_dts: List[date], **kws) -> ibis.expr.types.Table:
        """reads as ibis dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        filepaths = (self.get_file_path_with_protocol(snap_dt) for snap_dt in snap_dts)
        return self._ibis_con.read_parquet(filepaths, **kws)
    
    def read_pd(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_file_path(snap_dt=snap_dt)
        return pd.read_parquet(
            path = filepath,
            engine = 'pyarrow',
            filesystem = self._fs,
            **kws
        )
    
    def reads_pd(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
                # do the checks to ensure successful concatenation
        li =  []
        for snap_dt in snap_dts:
            df_partition = self.read_pd(snap_dt=snap_dt, **kws)
            li.append(df_partition)

        return pd.concat(li, axis = 0, ignore_index = True)
    
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        filepath = self.get_file_path(snap_dt=snap_dt)
        if self._fs.exists(filepath):
            self._fs.rm_file(path = filepath)

    def drop(self):
        """drop the whole table

        :return:
        """
        if self.exist():
            self._fs.rm(
                path = self.dir,
                recursive = True,
                maxdepth = None
            )

    def migrate(self, dst_schema: str, dst_table: str):
        """move from the existing schema.table to new one, the existing one will be deleted

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = self.duplicate(dst_schema, dst_table)
        self.drop()
        return new

    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerFileSystem:
        """duplicate the existing schema.table to new one, the existing one will be kept

        reference here: https://filesystem-spec.readthedocs.io/en/latest/copying.html
        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = type(self)(
            schema = dst_schema, 
            table = dst_table
        )
        # copy folder to new location
        self._fs.cp(
            path1 = str(Path(self.dir) / "_")[:-1],
            path2 = str(Path(new.dir) / "_")[:-1],
            recursive = True,
            maxdepth = None
        )
        

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date on disk

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        filepath = self.get_file_path(snap_dt=snap_dt)
        size_bytes = self._fs.du(
            path = filepath,
            total = True,
            maxdepth = None
        )
        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size
    
        
class SnapshotDataManagerWarehouseMixin(SnapshotDataManagerBase):
    """Mixin class for warehouse implementation (partial)
    """
    def __init__(self, schema:str, table:str, snap_dt_key: str):
        """database management interface

        :param schema: schema
        :param table:  table under each schema
        :param snap_dt_key: snap date column name for all tables
        """
        super().__init__(schema = schema, table = table)
        self.snap_dt_key = snap_dt_key
        
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        return self.is_exist(schema=self.schema, table=self.table)
    
    def init_table(self, col_schemas: DSchema, overwrite:bool = False, **settings):
        """initialize/create table in the underlying data warehouse system

        :param DSchema col_schemas: data column schema
        :param bool overwrite: whether to drop table if exists, defaults to False
        """

        if self.exist():
            if overwrite:
                self.drop()
            else:
                logging.warning(f"{self.schema}.{self.table} already exists, will do nothing." 
                                "set overwrite to True if you wish to reset table")
                return
        
        # create schema
        self.create_schema(schema = self.schema)
        # create table
        self.create_table(
            schema = self.schema,
            table = self.table,
            col_schemas = col_schemas,
            **settings
        )
    
    def get_schema(self) -> ibis.expr.schema.Schema:
        """Check dtypes of the given schema/dataset

        :return: the ibis schema of the given table
        """
        return self.get_dtypes(
            schema = self.schema,
            table = self.table
        )
    
    
    def get_existing_snap_dts(self) -> List[date]:
        """get existing snap dates as a list

        :return List[date]: list of snap dates
        """
        if not self.exist():
            return []
        existing_snaps = (
            self.get_table(
                schema = self.schema,
                table = self.table
            ).select(self.snap_dt_key)
            .distinct()
            .order_by(self.snap_dt_key)
            .to_pandas()
        )
        if len(existing_snaps) == 0:
            return []
        return list(
            str2dt(dt.date()) 
            for dt in pd.to_datetime(existing_snaps[self.snap_dt_key]).dt.to_pydatetime()
        )
    
    def read(self, snap_dt: date, **kws) -> ibis.expr.types.Table:
        """Read as ibis dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        table = self.get_table(
            schema = self.schema,
            table = self.table
        )
        df = table.filter(table[self.snap_dt_key] == snap_dt)
        if 'columns' in kws:
            df = df.select(*kws['columns'])
        return df
    
    def reads(self, snap_dts: List[date], **kws) -> ibis.expr.types.Table:
        """reads as ibis dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        table = self.get_table(
            schema = self.schema,
            table = self.table
        )
        df = table.filter(table[self.snap_dt_key].isin(snap_dts))
        if 'columns' in kws:
            df = df.select(*kws['columns'])
        return df
    
    def read_pd(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        return self.read(snap_dt=snap_dt, **kws).to_pandas()
    
    def reads_pd(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        return self.reads(snap_dts=snap_dts, **kws).to_pandas()

    def drop(self):
        """drop the whole table

        :return:
        """
        self.delete_table(
            schema = self.schema,
            table = self.table,
        )



def check_params(param, allowed_types: list = None, allowed_values: list = None):
    if allowed_types:
        if not isinstance(param, tuple(allowed_types)):
            raise TypeError(f"types allowed for {param}: {','.join(allowed_types)}")

    if allowed_values:
        if param not in allowed_values:
            raise ValueError(f"allowed values for {param}: {','.join(allowed_values)}")

    return param


import contextlib
import fnmatch
import os
from fsspec.implementations.sftp import SFTPFileSystem
from typing import List
import paramiko
import logging
from stat import S_ISDIR, S_ISREG
from tqdm.auto import tqdm


class SFTPFileSystem(SFTPFileSystem):
    """fsspec implementation of SFTP is not complete"""
    def cp_file(self, path1, path2, **kwargs):
        path1 = self._strip_protocol(path1)
        path2 = self._strip_protocol(path2)
        if self.isfile(path1):
            self.client.exec_command(f"cp {path1} {path2}")
        elif self.isdir(path1):
            self.mkdirs(path2, exist_ok=True)
        else:
            raise FileNotFoundError(path1)
        
class SFTP:
    def __init__(self, host: str, port: int, username: str, password: str, private_key_file: str=None):
        """Remote server operations using paramiko

        :param host: ip address of the remote server
        :param port: port number of the remote server
        :param username: username to login
        :param password: password to login, could be None if private key file is provided
        :param private_key_file: the path to the private key file (id_rsa, should be in openSSH format)

        Ref:
        https://avleonov.com/2017/09/05/ssh-sftp-public-key-authentication-and-python/
        https://stackoverflow.com/questions/8382847/how-to-ssh-connect-through-python-paramiko-with-ppk-public-key

        """
        self.host = host
        self.port = port
        self.username = username
        self.__password = password

        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.connect(host, port, username=username, password=password, key_filename=private_key_file)

    def bash(self, bash_cmd: str):
        """execute bash command on remote linux server

        :param bash_cmd: bash command string
        :return: stdin_, stdout_, stderr_
        """
        stdin_, stdout_, stderr_ = self.ssh.exec_command(bash_cmd)
        return stdin_, stdout_, stderr_

    def getFilePath(self, file: str):
        return f"SFTP://{self.username}@{self.host}:{self.port}/{file}"

    def ls(self, remoteFolderPath: str, pattern: str = None, verbose: bool = False) -> List[str]:
        """List files under a specific folder, and can specify filetype to filter

        :param remoteFolderPath: the remote folder path to list file
        :param pattern: the pattern regrex string for file filters, e.g., pattern = "*.csv" will only show all csv files
        :param verbose: whether to show the list
        :return:
        """
        with self.getSFTP() as sftp:
            file_list = sftp.listdir(remoteFolderPath)
            if pattern:
                file_list = list(filter(lambda x: fnmatch.fnmatch(x, pattern), file_list))
        if verbose:
            print(f"{len(file_list)} files found under {remoteFolderPath}:")
            for i, file in enumerate(file_list):
                print(i, file)
        return file_list

    def mkdir(self, path: str, mode: int = 511, ignore_existing: bool = False) -> int:
        """create a remote directory

        :param path: the remote directory path to create
        :param mode: mode of the file
        :param ignore_existing: if True, will not do anything when the remote folder already exists
        :return:
        """
        with self.getSFTP() as sftp:
            try:
                sftp.mkdir(path, mode)
            except IOError as e:
                if ignore_existing:
                    logging.info(f"Folder already exists, will not do anything")
                    return 1
                else:
                    logging.error(
                        f"Folder already exists: {path}, manually delete the folder or set ignore_existing to True")
                    return 0
            else:
                logging.info(f"Folder successfully created: {path}")
                return 1

    @contextlib.contextmanager
    def getSFTP(self):
        """Get an SFTP connector

        :return:
        """
        sftp = self.ssh.open_sftp()
        yield sftp
        sftp.close()

    @contextlib.contextmanager
    def getFileHandler(self, fullFilePath: str, mode: str = 'r'):
        """Get a file handler for remote file reader/writer

        :param fullFilePath: filepath, just like with open(filepath) as ...
        :param mode: mode of opening the file; r=read, w=write, a=append, etc.
        :return: the file context handler, can be accepted by pandas, json
        """
        with self.getSFTP() as sftp:
            fileObject = sftp.file(fullFilePath, mode)
            yield fileObject
            fileObject.close()

    def download(self, remoteFilePath: str, localFilePath: str):
        """download a file from remote server to local

        :param remoteFilePath:
        :param localFilePath:
        :return:
        """
        with self.getSFTP() as sftp:
            sftp.get(remoteFilePath, localFilePath)
            logging.info(f"Successfully download File from {self.getFilePath(remoteFilePath)} to {localFilePath}")

    def download_folder(self, remoteFolderPath: str, localFolderPath: str, pattern: str = None, verbose: bool = False):
        """download a remote folder to local

        :param remoteFolderPath:
        :param localFolderPath:
        :param pattern: the wildcard for file mask. e.g. *.csv to include only csv files
        :param verbose: if True, will print downloading details
        :return:

        Note: will create local folder if not exists
        """
        if not os.path.exists(localFolderPath):
            os.makedirs(localFolderPath)

        with self.getSFTP() as sftp:
            item_list = sftp.listdir_attr(remoteFolderPath)

            if not os.path.isdir(localFolderPath):
                os.makedirs(localFolderPath, exist_ok=True)

            for item in tqdm(item_list):
                remoteSubFolder = os.path.join(remoteFolderPath, item.filename)
                localSubFolder = os.path.join(localFolderPath, item.filename)

                if S_ISDIR(item.st_mode):
                    self.download_folder(remoteSubFolder, localSubFolder, pattern=pattern, verbose=verbose)
                else:
                    logging.info(f"Start downloading folder {remoteFolderPath} to {localFolderPath}")
                    if pattern:
                        if not fnmatch.fnmatch(item.filename, pattern):
                            continue
                    sftp.get(remoteSubFolder, localSubFolder)

                    if verbose:
                        logging.info(f"Successfully download File {item.filename}")

    def upload(self, localFilePath: str, remoteFilePath: str):
        """Upload a file from local to remote server

        :param localFilePath:
        :param remoteFilePath:
        :return:
        """
        with self.getSFTP() as sftp:
            sftp.put(localFilePath, remoteFilePath)
            logging.info(f"Successfully upload File from {localFilePath} to {self.getFilePath(remoteFilePath)}")

    def upload_folder(self, localFolderPath: str, remoteFolderPath: str, pattern: str = None, verbose: bool = False):
        """download a remote folder to local

        :param localFolderPath:
        :param remoteFolderPath:
        :param pattern: the wildcard for file mask. e.g. *.csv to include only csv files
        :param verbose: if True, will print uploading details
        :return:

        Note: will create remote folder if not exists
        """
        self.mkdir(remoteFolderPath, ignore_existing=True)

        with self.getSFTP() as sftp:
            for item in tqdm(os.listdir(localFolderPath)):
                remoteSubFolder = os.path.join(remoteFolderPath, item)
                localSubFolder = os.path.join(localFolderPath, item)

                if os.path.isfile(localSubFolder):
                    if pattern:
                        if not fnmatch.fnmatch(item, pattern):
                            continue
                    sftp.put(localSubFolder, remoteSubFolder)

                    if verbose:
                        logging.info(f"Successfully uploaded File {item}")
                else:
                    self.mkdir(remoteSubFolder, mode=511, ignore_existing=False)
                    self.upload_folder(localSubFolder, remoteSubFolder, pattern=pattern, verbose=verbose)

    def copy(self, from_path: str, to_path: str):
        cmd = f"cp {from_path} {to_path}"
        self.bash(cmd)
    
    
    def delete(self, remoteFolderPath: str):
        with self.getSFTP() as sftp:
            sftp.remove(remoteFolderPath)
    
    
    def delete_folder(self, remoteFolderPath: str):
        """Delete a folder recursively

        :param remoteFolderPath: the remote folder path to be deleted
        :return:
        """
        with self.getSFTP() as sftp:
            for item in tqdm(sftp.listdir_attr(remoteFolderPath)):
                rpath = os.path.join(remoteFolderPath, item.filename)

                if S_ISDIR(item.st_mode):
                    self.delete_folder(rpath)
                else:
                    sftp.remove(rpath)

            sftp.rmdir(remoteFolderPath)
            logging.info(f"Successfully delete folder: {remoteFolderPath}")
            
    def size(self, remoteFilePath: str) -> int:
        with self.getSFTP() as sftp:
            stat = sftp.lstat(remoteFilePath)
            if S_ISDIR(stat.st_mode):
                # if a folder
                size = 0
                for f in self.ls(remoteFilePath):
                    fpath = os.path.join(remoteFilePath, f)
                    size += self.size(fpath)
            else:
                size = stat.st_size
        return size
    
    def exist(self, path: str) -> bool:
        with self.getSFTP() as sftp:
            try:
                sftp.stat(path)
            except FileNotFoundError:
                return False
            else:
                return True


from typing import List, Literal, Tuple, Dict
import numpy as np
import pandas as pd
from luntaiDs.ModelingTools.Explore.summary import DescStat, QuantileStat, TabularStat, XtremeStat, \
    BaseStatAttr, BaseStatObj, BinaryStatAttr, BinaryStatSummary, CategStatAttr, CategStatSummary, \
        NominalCategStatAttr, NominalCategStatSummary, NumericStatAttr, \
        NumericStatSummary, OrdinalCategStatAttr, OrdinalCategStatSummary, \
        BinaryStatObj, OrdinalCategStatObj, NominalCategStatObj, NumericStatObj,\
        CategUniVarClfTargetCorr, NumericUniVarClfTargetCorr
from luntaiDs.ModelingTools.utils.parallel import delayer, parallel_run


def serialize(v):
    if pd.isnull(v):
        return None
    if v == np.inf:
        return float('inf')
    if v == -np.inf:
        return -float('inf')
    if hasattr(v, 'item'):
        return v.item()
    else:
        return v 
    
class _BaseNumericHelper:
    
    def get_descriptive_stat(self) -> DescStat:
        raise NotImplementedError("")
    
    def get_quantile_stat(self) -> QuantileStat:
        raise NotImplementedError("")        
    
    def get_xtreme_stat(self, xtreme_method: Literal["iqr", "quantile"] = "iqr") -> XtremeStat:
        raise NotImplementedError("")
        
    def get_histogram(self, n_bins:int) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError("")
    
class _BaseEDAEngine:
    def get_columns(self) -> List[str]:
        """get all column list from given dataset

        :return List[str]: list of columns in the dataset
        """
        raise NotImplementedError("")
        
    def _fit_common_categ(self, colname: str, attr: CategStatAttr) -> CategStatSummary:
        """common categorical variable fitting can be reused by subclass categorical fitting

        :param str colname: column name
        :param CategStatAttr attr: categorical variable attribute object
        :return CategStatSummary: categorical variable summary object
        """
        raise NotImplementedError("")
        
    def fit_binary(self, colname: str, attr: BinaryStatAttr) -> BinaryStatSummary:
        """binary categorical variable fitting

        :param str colname: column name
        :param BinaryStatAttr attr: binary variable attribute object
        :return BinaryStatSummary: binary variable summary object
        """
        raise NotImplementedError("")
    
    def fit_binary_obj(self, colname: str, attr: BinaryStatAttr) -> BinaryStatObj:
        """binary categorical variable fitting

        :param str colname: column name
        :param BinaryStatAttr attr: binary variable attribute object
        :return BinaryStatObj: binary variable result object (attr + summary)
        """
        return BinaryStatObj(
            colname = colname,
            attr = attr,
            summary = self.fit_binary(colname, attr)
        )
        
    def fit_ordinal(self, colname: str, attr: OrdinalCategStatAttr) -> OrdinalCategStatSummary:
        """ordinal categorical variable fitting

        :param str colname: column name
        :param OrdinalCategStatAttr attr: ordinal variable attribute object
        :return OrdinalCategStatSummary: ordinal variable summary object
        """
        # reuse common stat
        common_stat_summary = self._fit_common_categ(colname, attr)
        return OrdinalCategStatSummary(
            colname_ = common_stat_summary.colname_,
            total_ = common_stat_summary.total_,
            missing_ = common_stat_summary.missing_,
            unique_ = common_stat_summary.unique_,
            vcounts_ = common_stat_summary.vcounts_,
            vpercs_ = common_stat_summary.vpercs_,
        )
        
    def fit_ordinal_obj(self, colname: str, attr: OrdinalCategStatAttr) -> OrdinalCategStatObj:
        """ordinal categorical variable fitting

        :param str colname: column name
        :param OrdinalCategStatAttr attr: ordinal variable attribute object
        :return OrdinalCategStatObj: ordinal variable result object (attr + summary)
        """
        return OrdinalCategStatObj(
            colname = colname,
            attr = attr,
            summary = self.fit_ordinal(colname, attr)
        )
        
    def fit_nominal(self, colname: str, attr: NominalCategStatAttr) -> NominalCategStatSummary:
        """nominal categorical variable fitting

        :param str colname: column name
        :param NominalCategStatAttr attr: nominal variable attribute object
        :return NominalCategStatSummary: nominal variable summary object
        """
        # reuse common stat
        common_stat_summary = self._fit_common_categ(colname, attr)
        return NominalCategStatSummary(
            colname_ = common_stat_summary.colname_,
            total_ = common_stat_summary.total_,
            missing_ = common_stat_summary.missing_,
            unique_ = common_stat_summary.unique_,
            vcounts_ = common_stat_summary.vcounts_,
            vpercs_ = common_stat_summary.vpercs_,
        )
        
    def fit_nominal_obj(self, colname: str, attr: NominalCategStatAttr) -> NominalCategStatObj:
        """nominal categorical variable fitting

        :param str colname: column name
        :param NominalCategStatAttr attr: nominal variable attribute object
        :return NominalCategStatObj: nominal variable result object (attr + summary)
        """
        return NominalCategStatObj(
            colname = colname,
            attr = attr,
            summary = self.fit_nominal(colname, attr)
        )
        
    def fit_numeric(self, colname: str, attr: NumericStatAttr) -> NumericStatSummary:
        """numeric variable fitting

        :param str colname: column name
        :param NumericStatAttr attr: numeric variable attribute object
        :return NumericStatSummary: numeric variable summary object
        """
        raise NotImplementedError("")
    
    def fit_numeric_obj(self, colname: str, attr: NumericStatAttr) -> NumericStatObj:
        """numeric variable fitting

        :param str colname: _description_
        :param NumericStatAttr attr: numeric variable attribute object
        :return NumericStatObj: numeric variable result object (attr + summary)
        """
        return NumericStatObj(
            colname = colname,
            attr = attr,
            summary = self.fit_numeric(colname, attr)
        )
        
    
    @delayer
    def _fit_one(self, col: str, attr: BaseStatAttr) -> BaseStatObj:
        if isinstance(attr, NumericStatAttr):
            return self.fit_numeric_obj(col, attr)
        if isinstance(attr, NominalCategStatAttr):
            return self.fit_nominal_obj(col, attr)
        if isinstance(attr, OrdinalCategStatAttr):
            return self.fit_ordinal_obj(col, attr)
        if isinstance(attr, BinaryStatAttr):
            return self.fit_binary_obj(col, attr)
        else:
            raise TypeError("Only [NumericStatAttr/NominalCategStatAttr/OrdinalCategStatAttr/BinaryStatAttr] are supported")
    
    def fit(self, attrs: Dict[str, BaseStatAttr], n_jobs: int = 1) -> TabularStat:
        """fit a tabular dataset

        :param Dict[str, BaseStatAttr] attrs: dictionary of col:attribute pairs
        :param int n_jobs: control parallelism, defaults to 1
        :return TabularStat: tabular stat object
        """
        df_cols = self.get_columns()
        used_cols = [col for col in attrs.keys() if col in df_cols]
        jobs = (
            self._fit_one(col, attrs.get(col)) 
            for col in used_cols
        )
        summaries = parallel_run(jobs, n_jobs = n_jobs)
        return TabularStat(zip(used_cols, summaries))
    

    def fit_univarclf_categ(self, x_col: str, y_col: str) -> CategUniVarClfTargetCorr:
        """fit univariate correlation for classififer target and categorical variable

        :param str x_col: column of a categorical variable
        :param str y_col: column of the classifier target varaible
        :return CategUniVarClfTargetCorr: result object
        """
        raise NotImplementedError("")
    
    def fit_univarclf_numeric(self, x_col: str, y_col: str) -> NumericUniVarClfTargetCorr:
        """fit univariate correlation for classififer target and numeric variable

        :param str x_col: column of a numeric variable
        :param str y_col: column of the classifier target varaible
        :return NumericUniVarClfTargetCorr: result object
        """
        raise NotImplementedError("")



from __future__ import annotations
from datetime import date
import logging
from typing import List
import ibis
from ibis import _
from fsspec.implementations.local import LocalFileSystem
from fsspec import AbstractFileSystem
from luntaiDs.CommonTools.SnapStructure.structure import SnapshotDataManagerWarehouseMixin
from luntaiDs.ProviderTools.clickhouse.dbapi import ClickHouse, WarehouseHandlerCHSQL

class SnapshotDataManagerCHSQL(SnapshotDataManagerWarehouseMixin, WarehouseHandlerCHSQL):
    """files are saved as clickhouse tables under each schema.table
    """
    @classmethod
    def setup(cls, db_conf: ClickHouse, **settings):
        super(SnapshotDataManagerCHSQL, cls).setup()
        cls.connect(db_conf=db_conf, **settings)
        
    def save_qry(self, query: str, snap_dt: date, overwrite: bool = False):
        """save the query into the table using "insert into schema.table select ... " clause

        :param query: the sql query for table to be inserted
        :param snap_dt: the snap date partition to be inserted, use to check existence
        :param overwrite: whether to overwrite the table if exists

        :return:
        """
        # first detect whether snap date already exists
        clear = self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite)
        if clear is False:
            return

        # first create a view of that table which can be inspected the column dtypes and names
        qry_cols = self.query(query).columns

        # insert into the table
        # clickhouse insert need column order correct
        sql = f"""
        INSERT INTO {self.schema}.{self.table} ({','.join(qry_cols)})
        {query}
        """
        logging.info(f"Inserting into {self.schema}.{self.table}@{snap_dt} using query:\n{sql}")
        self.execute(sql)

        logging.info(f"Successfully saved to {self.schema}.{self.table}@{snap_dt}")
        
    def _save(self, df: ibis.expr.types.Table, snap_dt: date, **kws):
        """The pure logic to save ibis dataframe to the system, without handling existing record problem

        :param ibis.expr.types.Table df: the ibis dataframe
        :param date snap_dt: the snap date to save
        """
        self.save_ibis(
            df = df,
            schema = self.schema,
            table = self.table,
            **kws
        )
        
    def ingest_from_file(self, snap_dt: date, file_path: str, 
                        fs: AbstractFileSystem = LocalFileSystem(), 
                        header:bool = True, overwrite:bool = True, 
                        column_names: List[str] = None, compression: str = None, **settings):
        """ingest from a parquet file from any fsspec compatible filesystem

        :param date snap_dt: snap date for the file to ingest
        :param str file_path: the absolute filepath on the given filesystem
        :param AbstractFileSystem fs: fsspec compatible FS, defaults to LocalFileSystem()
        :param bool header: where have a header or not, defaults to True
        :param bool overwrite: whether to overwrite the existing data, defaults to True
        :param List[str] column_names: used in clickhouse raw insert, defaults to None
        :param str compression: compression format, defaults to None
        """
        if overwrite:
            self.delete(snap_dt = snap_dt)
        else:
            num_records = self.count(snap_dt)
            if num_records > 0:
                logging.warning(f"{num_records} records found for "
                                f"{self.schema}.{self.table}@{snap_dt}, will do nothing")
                return
            
        # extract file format
        if file_path.endswith("parquet"):
            fmt = "Parquet"
        elif file_path.endswith("csv"):
            fmt = "CSVWithNames" if header else "CSV"
        else:
            raise TypeError("File not supported")
        
        from clickhouse_connect.driver.query import quote_identifier

        if not self.schema and self.table[0] not in ('`', "'") and self.table.find('.') > 0:
            full_table = self.table
        elif self.schema:
            full_table = f'{quote_identifier(self.schema)}.{quote_identifier(self.table)}'
        else:
            full_table = quote_identifier(self.table)
        if not fmt:
            fmt = 'CSVWithNames'
        if compression is None:
            if file_path.endswith('.gzip') or file_path.endswith('.gz'):
                compression = 'gzip'
        with fs.open(file_path, 'rb') as obj:
            self._ops.con.raw_insert(
                full_table,
                column_names=column_names,
                insert_block=obj,
                fmt=fmt,
                settings=settings,
                compression=compression
            )
        logging.info(f"Successfully ingested file {file_path} to {self.schema}.{self.table}")

        
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        if self.exist():
            sql = f"""
            ALTER TABLE {self.schema}.{self.table} DELETE
            WHERE {self.snap_dt_key} = '{snap_dt}'
            """
            logging.info(f"Deleting table {self.schema}.{self.table} using query:\n{sql}")
            self.execute(sql = sql)
        
    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerCHSQL:
        """duplicate the existing schema.table to new one, the existing one will be kept

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        args = dict(
            dst_schema = dst_schema, 
            dst_table = dst_table, 
            src_schema = self.schema, 
            src_table = self.table
        )
        # create table structure
        sql = f"""
        create table if not exists {dst_schema}.{dst_table}
        AS {self.schema}.{self.table}"""
        self.execute(sql = sql)
        
        # insert from select
        sql = """
        insert into {dst_schema:Identifier}.{dst_table:Identifier} 
        select * from {src_schema:Identifier}.{src_table:Identifier}"""
        self.execute(sql = sql, parameters = args)
        new = SnapshotDataManagerCHSQL(
            schema = dst_schema,
            table = dst_table,
            snap_dt_key = self.snap_dt_key
        )
        return new
    
    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date file (pandas) or folder (pyspark partitions)

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        sql = """
        select 
            sum(rows) as rows,
            sum(bytes_on_disk) as bytes_on_disk
        from system.parts
        where 
            active
            and database = %(schema)s
            and table = %(table)s
            and partition = %(snap_dt)s
        """
        if not self.exist():
            return 0
        
        args = dict(schema = self.schema, table =self.table, snap_dt = snap_dt)
        d = self._ops.con.query(query=sql, parameters=args).first_item

        size_bytes, rows = d['bytes_on_disk'], d['rows']
        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size


from __future__ import annotations
from datetime import date
import logging
import time
from typing import List, Literal
from collections import OrderedDict
from luntaiDs.CommonTools.SnapStructure.structure import SnapshotDataManagerBase
from luntaiDs.CommonTools.SnapStructure.tools import get_future_period_ends, get_past_period_ends
from luntaiDs.CommonTools.utils import str2dt, offset_date

""" Chain Executor with dependency check """

class _CurrentStream:
    def __init__(self, upstream, offset: int = 0, 
                 freq: Literal['M', 'D', 'W'] = 'M'):
        self.upstream = upstream
        self.offset = offset
        self.freq = freq

    def get_exc_dts(self, snap_dt: date):
        if self.offset != 0:
            snap_dt = offset_date(snap_dt, offset = self.offset, freq = self.freq)
        return [snap_dt]


class _FutureStream(_CurrentStream):
    def __init__(self, upstream, future: int, offset: int = 0, 
                 freq: Literal['M', 'D', 'W'] = 'M'):
        super().__init__(upstream)
        self.future = future
        self.freq = freq
        self.offset = offset

    def get_exc_dts(self, snap_dt: date):
        if self.offset != 0:
            snap_dt = offset_date(snap_dt, offset = self.offset, freq = self.freq)
        return get_future_period_ends(snap_dt, forward=self.future, freq = self.freq)


class _PastStream(_CurrentStream):
    def __init__(self, upstream, history: int, offset: int = 0, 
                 freq: Literal['M', 'D', 'W'] = 'M'):
        super().__init__(upstream)
        self.history = history
        self.freq = freq
        self.offset = offset

    def get_exc_dts(self, snap_dt: date):
        if self.offset != 0:
            snap_dt = offset_date(snap_dt, offset = self.offset, freq = self.freq)
        return get_past_period_ends(snap_dt, history=self.history, freq = self.freq)[::-1]
    

class ExecPlan:
    def __init__(self, obj, snap_dt: date):
        self.obj = obj
        self.snap_dt = snap_dt
        self.upstreams = []

    def __str__(self):
        return f"{self.obj}@{self.snap_dt} | upstreams = {len(self.upstreams)}"

    def attach_upstream(self, plan: ExecPlan):
        self.upstreams.append(plan)

    def exec_plan(self) -> dict:
        up_infos = []
        for up in self.upstreams:
            up_infos.append(up.exec_plan())

        return {
            'current': self,
            'upstream' : up_infos
        }
    
    def get_serial_steps(self) -> list:
        """get a serias of steps to execute

        :return list: _description_
        """
        r = []
        for up in self.upstreams:
            r.extend(up.get_serial_steps())
        r.append(self)
        return r
    
    def reduce_summary(self) -> OrderedDict:
        """combine similar steps (same obj and snap date) from serial plan and do summarize

        :return dict: _description_
        """
        summary = OrderedDict()
        steps = self.get_serial_steps()
        for step in steps:
            if step.obj in summary.keys():
                if step.snap_dt not in summary[step.obj]:
                    summary[step.obj].append(step.snap_dt)
            else:
                summary[step.obj] = [step.snap_dt]
        return summary



class SnapTableStreamGenerator:
    dm: SnapshotDataManagerBase = None
    upstreams: List[_CurrentStream] = []
    
    @classmethod
    def init(cls, snap_dt: date):
        """do some initialization if not exist or ready

        :param date snap_dt: _description_
        """
        pass

    @classmethod
    def execute(cls, snap_dt: date):
        """key logic for generating data at given snap date

        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError("")
    
    @classmethod
    def if_success(cls, snap_dt: date) -> bool:
        """determine if the run is successful or not

        :param date snap_dt: _description_
        :return bool: if True will run next job/upstream, otherwise will delete and rerun
        """
        return cls.dm.is_valid(snap_dt = snap_dt)

    @classmethod
    def _execute(cls, snap_dt: date):
        """execute the excute function defined

        :param date snap_dt: _description_
        """
        logging.info(f"Running {cls.dm}@{snap_dt}")
        snap_dt = str2dt(snap_dt)
        # single run, without checking the upstream tables

        start = time.time()
        cls.execute(snap_dt = snap_dt)
        end = time.time()
        seconds = end - start
        logging.info(f"Saved to {cls.dm}@{snap_dt}, used time {seconds // 60 : .0f} minutes {seconds % 60 : .0f} seconds")

    
    @classmethod
    def _prerun(cls, snap_dt: date):
        """preparation before actual run
        

        :param date snap_dt: _description_
        :return _type_: _description_
        """
        if cls.dm.exist():
            if snap_dt in cls.dm.get_existing_snap_dts():
                if cls.if_success(snap_dt):
                    logging.info(f"Already extracted for {cls.dm}@{snap_dt}, will bypass")
                    return
                else:
                    logging.warning(f"Existing records not valid for {cls.dm}@{snap_dt}, will rerun")
                    cls.dm.delete(snap_dt)
        else:
            logging.info(f"Table-Schema does not exist, will initialize")
            cls.init(snap_dt)
    
    @classmethod
    def run(cls, snap_dt: date):
        """run current and upstreams

        :param date snap_dt: _description_
        """
        snap_dt = str2dt(snap_dt)
        # prepare
        cls._prerun(snap_dt)

        # check upstream table recursively and make up any missing files
        for up in cls.upstreams:
            logging.info(f"Upstream found: {up.upstream.dm}")
            for dt in up.get_exc_dts(snap_dt):
                logging.info(f"Run upstream {up.upstream.dm} @ {dt}")
                up.upstream.run(dt)

        # execute
        cls._execute(snap_dt)
        
    @classmethod
    def run_detached(cls, snap_dt: date):
        """run current timestamp only, ignore upstreams

        :param date snap_dt: _description_
        :return _type_: _description_
        """
        snap_dt = str2dt(snap_dt)
        # prepare
        cls._prerun(snap_dt)
            
        # execute
        cls._execute(snap_dt)
            

    @classmethod
    def get_exec_plan(cls, snap_dt: date) -> ExecPlan:
        """get the execution plan object for further analysis

        :param date snap_dt: _description_
        :return ExecPlan: _description_
        """
        snap_dt = str2dt(snap_dt)

        plan = ExecPlan(
            obj = cls,
            snap_dt = snap_dt
        )

        for up in cls.upstreams:
            for dt in up.get_exc_dts(snap_dt):
                plan.attach_upstream(up.upstream.get_exec_plan(dt))

        return plan


from __future__ import annotations
import json
from pathlib import Path
from typing import List, Dict
from fsspec import AbstractFileSystem
import pandas as pd
from datetime import datetime

class TimeInterval:
    def __init__(self, start: datetime, end: datetime):
        if start >= end:
            raise ValueError("Start must be ealier than end datetime")
        self.start = start
        self.end = end

    def __str__(self) -> str:
        return f"TimeInterval({self.start} - {self.end})"
    
    def __contains__(self, dt: datetime) -> bool:
        return self.start <= dt <= self.end

    def overlap(self, ti: TimeInterval) -> bool:
        return ti.end > self.start and self.end > ti.start
    
    def overlap_type(self, ti: TimeInterval) -> int:
        """judge which type is the overlapping

        there are five types
        0 - no overlapping
        1 - ti is left of self
        2 - ti is right of self
        3 - ti includes self
        4 - self includes ti
        5 - perfect overlap

        :param TimeInterval ti: _description_
        :return int: {0,1,2,3,4}
        """
        if not self.overlap(ti):
            return 0
        elif  ti.start == self.start and self.end == ti.end:
            return 5
        elif ti.start < self.start < ti.end < self.end:
            return 1
        elif self.start < ti.start < self.end < ti.end:
            return 2
        elif ti.start <= self.start < self.end <= ti.end:
            return 3
        elif self.start <= ti.start < ti.end <= self.end:
            return 4

    def to_js(self) -> Dict[str, str]:
        return {
            'start' : self.start.isoformat(), 
            'end' : self.end.isoformat()
        }
    
    @classmethod
    def from_js(cls, l: Dict[str, str]) -> TimeInterval:
        return TimeInterval(
            start = datetime.fromisoformat(l['start']),
            end = datetime.fromisoformat(l['end'])
        )


class _BaseModelTimeTable:
    """you can have different model (model_id) run in different time period as prod model
    support for multiple schedules
    """
    def load_tb(self) -> Dict[str, List[TimeInterval]]:
        """load time table

        :return Dict[str, List[TimeInterval]]: time table loaded
        """
        raise NotImplementedError("")
    
    def save_tb(self, tb: Dict[str, List[TimeInterval]]):
        """save timetable

        :param Dict[str, List[TimeInterval]] tb: time table
        """
        raise NotImplementedError("")

    @ classmethod
    def sort_and_merge(cls, timetable: Dict[str, List[TimeInterval]]) -> Dict[str, List[TimeInterval]]:
        """sort and merge timetable to avoid any overlapping conflict

        :param Dict[str, List[TimeInterval]] timetable: timetable to be sorted
        :return Dict[str, List[TimeInterval]]: sorted and merged timetable
        """
        r = {}
        for mi, tis in timetable.items():
            length = len(tis)
            if length == 1:
                r[mi] = tis
                continue

            sort_tis = sorted(tis, key=lambda x: x.start)
            s = []; i = 0
            while i < length:
                cur = sort_tis[i]
                if i + 1 < length:
                    nex = sort_tis[i + 1]
                    if cur.end == nex.start:
                        cur.end = nex.end
                        i += 1
                s.append(cur)
                i += 1

            r[mi] = s
        return r


    def register(self, model_id: str, start: datetime, end: datetime,  force: bool = False):
        """add a new time interval into the schedule

        :param str model_id: model id to be registered
        :param datetime start: start datetime of this model id
        :param datetime end: end datetime of this model id
        :param bool force: if True, will move endpoints of existing periods to avoid overlapping
        """
        new = TimeInterval(start=start, end=end)
        timetable = self.load_tb()
        for mi, tis in timetable.items():
            del_list = []
            for i, ti in enumerate(tis.copy()):
                if new.overlap(ti):
                    if force is False:
                        raise ValueError(f"Overlap with existing model {mi} serving time {ti}, set force to True or change endpoints")
                    else:
                        ti_end = ti.end
                        ti_start = ti.start
                        ovt = new.overlap_type(ti)
                        if ovt == 1:
                            # end of ti -> start of new
                            timetable[mi][i].end = start
                        elif ovt == 2:
                            # start of ti -> end of new
                            timetable[mi][i].start = end
                        elif ovt == 3:
                            # end of ti -> start of new 
                            # + create new interval [end of new, end of ti]
                            timetable[mi][i].end = start
                            timetable[mi].append(TimeInterval(start=end, end=ti_end))
                        elif ovt == 4 or ovt == 5:
                            # remove ti
                            del_list.append(i)
                            # del timetable[mi][i]  # will cause index problem when deleting
            
            # delete task
            cleaned = [timetable[mi][i] for i in range(len(timetable[mi])) if i not in del_list]
            timetable[mi] = cleaned

        
        if model_id in timetable:
            timetable[model_id].append(new)
        else:
            timetable[model_id] = [new]
        
        timetable = self.sort_and_merge(timetable)
        self.save_tb(timetable)

    def get_model_id_by_datetime(self, dt: datetime) -> str:
        """get model id given run datetime

        :param datetime dt: run datetime
        :return str: model id in place
        """
        timetable = self.load_tb()
        for mi, tis in timetable.items():
            for ti in tis:
                if dt in ti:
                    return mi
                
    def get_schedule(self) -> pd.DataFrame:
        """get runnning timetable

        :return pd.DataFrame: timetable in pandas
        """
        timetable = self.load_tb()
        
        r =[]
        for mi, tis in timetable.items():
            for ti in tis:
                t = ti.to_js()
                t['model_id'] = mi
                r.append(t)
        
        df = pd.DataFrame.from_records(r)
        if len(df) > 0:
            df['start'] = pd.to_datetime(df['start'])
            df['end'] = pd.to_datetime(df['end'])
            df.sort_values(by = 'start', inplace=True, ignore_index=True)

        return df
    
class ModelTimeTableFileSystem(_BaseModelTimeTable):
    """you can have different model (model_id) run in different time period as prod model
    support for multiple schedules
    """
    def __init__(self, fs: AbstractFileSystem, tb_js_path: str):
        """model time table implemented for any fsspec compatible filesystem

        :param AbstractFileSystem fs: the fsspec compatible filesystem
        :param str tb_js_path: path of the js config file, if on object storage, 
            the full path including buckets
        """
        self._fs = fs
        self._tb_js_path = tb_js_path
        self._fs.makedirs(
            path = Path(tb_js_path).parent.as_posix(),
            exist_ok = True
        )

    def load_tb(self) -> Dict[str, List[TimeInterval]]:
        """load time table

        :return Dict[str, List[TimeInterval]]: time table loaded
        """
        tb = {}
        if self._fs.exists(self._tb_js_path):
            with self._fs.open(self._tb_js_path, 'r') as obj:
                c = json.loads(obj.read())
            
            for model_id, intervals in c.items():
                tb[model_id] = [TimeInterval.from_js(interval) for interval in intervals]
        return tb
    
    def save_tb(self, tb: Dict[str, List[TimeInterval]]):
        """save timetable

        :param Dict[str, List[TimeInterval]] tb: time table
        """
        c = {}
        for model_id, intervals in tb.items():
            c[model_id] = [interval.to_js() for interval in intervals]
            
        with self._fs.open(self._tb_js_path, 'w') as obj:
            json.dump(c, obj, indent = 4)